/* ------------------------------------------------------------------------- */
/*  Linux driver for MPEG-2 encoder based on Kfir chip.
    Written by Ralph Metzler for
    convergence integrated media gmbh, http://www.convergence.de

    for details see http://linuxtv.org/mpeg2
     

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                */
/* ------------------------------------------------------------------------- */


#define __KERNEL_SYSCALLS__
#include <linux/poll.h>
#include <linux/fs.h>
#include <linux/unistd.h>
#include <linux/fcntl.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/wrapper.h>
#include <linux/init.h>
#include <linux/ctype.h>
#include <asm/io.h>
#if defined(MODVERSIONS)
#include <linux/modversions.h>
#endif

#include "videodev.h"
#include "i2c.h"
#include "i2c-algo-bit.h"
#include "video_decoder.h"

#include "micro/kfir_micro.h"
#include "kfir.h"
#include "plx9054.h"

MODULE_AUTHOR("Ralph Metzler <ralph@convergence.de>");
MODULE_DESCRIPTION("Visiontech Kfir MPEG2 encoder driver");
MODULE_PARM(debug,"i");
MODULE_PARM(loaded,"i");
MODULE_PARM(vidinmode,"i");
MODULE_PARM(vidoutmode,"i");
MODULE_PARM(vidsize,"i");
MODULE_PARM(vidrate,"i");
MODULE_PARM(mpeg1mode,"i");
MODULE_PARM(streamtype,"i");
MODULE_PARM(sndfreq,"i");

static int debug = 0;
static int loaded = 0;
static int vidinmode = VIDEO_MODE_PAL;
static int vidoutmode = VT_KFIR_PAL;
static int vidsize = (KFIR_V_576*16+KFIR_H_FD1);
static int vidrate = 4500000;
static int sndfreq = VT_KFIR_AUDIO_FREQ_48_000;
static int mpeg1mode = 0;
static int streamtype = VT_KFIR_PROGRAM;


#define dprintk	if (debug) printk

static struct kfir_dev kfirs[MAX_NUM_KFIR];
static int num_kfir = 0;

// Define to allow burst mode with memory mapping
//#define ALTERA_BURST  1
#define PLX_BURST   1


/* Not a full fletched 64 bit routine but should be accurate enough for what we need */
inline u64 div64(u32 a, u32 b, u32 m)
{
	u32 r, s;
	
	if (!b)
		return 0;
	r=(a/b);
	s=((a%b)*m)/b;
	
	return (((u64 )r)*m)|s;
}

#define RMUL(a,b,c) div64((a),(c),(b))


static inline void ddelay(int i) 
{
        current->state=TASK_INTERRUPTIBLE;
        schedule_timeout((HZ*i)/100);
}

static inline void Sleep(int i) 
{
	if (i>10)
		ddelay(1+i/10);
	else
		mdelay(i);
}

static inline
void AudioRegWrite(struct kfir_dev *kfir, WORD const addr , WORD const data)
{
	switch( addr ) {
	case IAL:
		waltl(ALT_ADR_ADSP_ADDR , data);
		break;
	case IDMA:
		waltl(ALT_ADR_ADSP_DATA, data);
		break;
	case GPPORT:
		waltl(ALT_ADR_ADSP_IRQ, data);
		break;
	default:
		break;
	}
}

static inline
WORD AudioRegRead(struct kfir_dev *kfir, WORD const addr)  
{
	switch( addr ) {
	case IAL:
		return raltl(ALT_ADR_ADSP_ADDR);
		break;
	case IDMA:
		return raltl(ALT_ADR_ADSP_DATA);
		break;
	case GPPORT:
	default:
		return 0;
	}
}

static int errno;
static int do_firmread(const char *fn, char **fp)
{
        int fd;
	long l;
	char *dp;
        
	fd = open(fn, 0, 0);
	if (fd == -1)
	{
		printk(KERN_INFO "Unable to load '%s'.\n", fn);
		return 0;
	}
	l = lseek(fd, 0L, 2);
	if (l <= 0 || l > 300*1024)
	{
	        printk(KERN_INFO "Firmware too large '%s'\n", fn);
		sys_close(fd);
		return 0;
	}
	lseek(fd, 0L, 0);
	*fp= dp = vmalloc(l);
	if (dp == NULL)
	{
		printk(KERN_INFO "Out of memory loading '%s'.\n", fn);
		sys_close(fd);
		return 0;
	}
	if (read(fd, dp, l) != l)
	{
		printk(KERN_INFO "Failed to read '%s'.\n", fn);
		vfree(dp);
		sys_close(fd);
		return 0;
	}
	close(fd);
	*fp = dp;
	return (int) l;
}

static
int firmread(const char *fn, char **fp)
{
	int r;
	mm_segment_t fs = get_fs();
        
	set_fs(get_ds());
	r = do_firmread(fn, fp);
	set_fs(fs);
	return r;
}



/* ----------------------------------------------------------------------- */
/* ADSP functions                                                          */

static inline void ADSP_xsleep2(int i) 
{
	udelay(i*100);
}

static u16 ADSP_MonoRates[10]   = {32,48,56,64,80,96,112,128,160,192};
static u16 ADSP_StereoRates[10] = {64,96,112,128,160,192,224,256,320,384};
static u16 ADSP_AllRates[14]    = {32,48,56,64,80,96,112,128,160,192,224,256,320,384};
static u16 ADSP_FreqSample[8]   = {0x500,0x700,0x400,0x600,0x300,0x200,0x000,0x100};

static const WORD* ADSP_GetRates(struct kfir_dev *kfir, WORD AudioMode, 
				 WORD AudioFormat)
{
        if (!AudioFormat)
	        return ADSP_AllRates;
	else
	        if (!AudioMode)
		        return ADSP_MonoRates;
		else
		        return ADSP_StereoRates;
}


static void ADSP_SendSystemCommand(struct kfir_dev *kfir, const WORD Command)
{
	switch(Command) {
	case SYSCMD_CANCELMASTERINT :
		AudioRegWrite(kfir, GPPORT, kfir->m_CmdReg & CLEARIRQLED);
		break;
		
	case SYSCMD_MASTERDSPINT:
	        //AudioRegWrite(kfir, GPPORT, CmdReg & RAISEIRQ2INT);
		AudioRegWrite(kfir, GPPORT , kfir->m_CmdReg & ~ALT_D_ADSP_IRQ_2);
		break;
		
	case SYSCMD_RAISEIRQEINT:
		//AudioRegWrite(kfir, GPPORT, CmdReg & RAISEIRQEINT);
		AudioRegWrite(kfir, GPPORT , kfir->m_CmdReg & ~ALT_D_ADSP_IRQ_2);
		break;
		
	default:
		if(Command & SYSCMD_STARTMASTERDSP)
			kfir->m_CmdReg |= START_DSP;
		if(Command & SYSCMD_STOPMASTERDSP)
			kfir->m_CmdReg &=  ~START_DSP;
	}
	
	AudioRegWrite(kfir, GPPORT, kfir->m_CmdReg);
	return;
}

static WORD SendDSPApiCommand(struct kfir_dev *kfir, const WORD Command )
{
	u32 start;
	WORD wd=0x8000;
	
	AudioRegWrite(kfir, IAL, ADR4);
	AudioRegWrite(kfir, IDMA, 0xffff);

	AudioRegWrite(kfir, IAL, ADR0);
	ADSP_xsleep2(10);
	AudioRegWrite(kfir, IDMA, Command);
        ADSP_xsleep2(5);
	waltl(ALT_ADR_ADSP_IRQ, 0);
	udelay(10);
	waltl(ALT_ADR_ADSP_IRQ, 2);
	//SP_SendSystemCommand(kfir, SYSCMD_MASTERDSPINT);

	start=jiffies;
	while(jiffies-start <= HZ) {  
		Sleep(3);
		AudioRegWrite(kfir, IAL, ADR4);
		wd=AudioRegRead(kfir, IDMA);
		if (wd != 0xFFFF) {
			AudioRegWrite(kfir, IAL, ADR4);
			wd=AudioRegRead(kfir, IDMA);
			break;
		}
	}
	
	if ((wd & 0x8000)==0)
		return wd;
	
	return 0x8000;
}


#define PLL_DELAY   180
static int ADSP_SetupPLL(struct kfir_dev *kfir,  const WORD AudioFreq )
{
	DWORD Data, Mask;
	int i,j;
	
	switch( kfir->revision ) {
	case KFIR_BOARD_A:
	case KFIR_BOARD_B:
		Data=raltl( ALT_ADR_ADSP_FREQ);
		Data &= ~3;
		Data |= 0xC;
		
		switch ( AudioFreq ) {
		case 0: // return "5.5125KHz";
		case 1: // return "8KHz";
		case 2: // return "11.025KHz";
		case 3: // return "16KHz";
		case 4: // return "22.5KHz";
		default:
			//ASSERT( 0 );
		case 5: // return "32KHz"; // Trick !!!
			break;
		case 6: // return "44.1KHz";
			Data |= 1;
			break;
		case 7: // return "48KHz";
			Data |= 2;
		}
		waltl(ALT_ADR_ADSP_FREQ, Data );
		break;
	case KFIR_BOARD_C:
	case KFIR_BOARD_D:
		{
			DWORD Address = PLX9054_CNTRL_OFFSET;
			DWORD PllData = 0x1AFF7FC;
			int r , x , v , m , n;
			
			switch ( AudioFreq ) {
			case VT_KFIR_AUDIO_FREQ_5_5125:
			case VT_KFIR_AUDIO_FREQ_8_0000:
			case VT_KFIR_AUDIO_FREQ_11_025:
			case VT_KFIR_AUDIO_FREQ_16_000:
			case VT_KFIR_AUDIO_FREQ_22_050:
				break;
			case VT_KFIR_AUDIO_FREQ_32_000:
				PllData = 0x1AFD7A3;
				r = 3; x = 0; v =0 ; m = 32 ; n = 127;
				break;
			case VT_KFIR_AUDIO_FREQ_44_100:
				PllData = 0x1AFD327;
				r = 3; x = 0; v =0 ; m = 19 ; n = 104;
				break;
			default:
			case VT_KFIR_AUDIO_FREQ_48_000:
				PllData = 0x1AFF7FC;
				r = 3; x = 0; v = 1 ; m = 47 ; n = 35;
				break;
			}
			// PllData = (1<<24) | (1 << 23) | (0 << 22) | (1<<21) |(1<<19 ) | (r<<17) | (x<<15) | (v<<14) | (m<<7) | n;
			printk( "kfir: PllData = %X\n" , PllData);
			Data=rplxl(Address);
			// Enable User Output
			Data |= GP_OUTPUT_ENABLE;
			Data=(Address);
			// Disable CS of the EEPROM
			Data &= ~SERIAL_EEPROM_CS;
			wplxl(Address, Data);
			// PLL CE = 1
			Data |= GP_OUTPUT;
			wplxl(Address, Data);
			for ( j = 0 ; j < PLL_DELAY ; j++ ) ; // delay
			// PLL CLK = 0
			Data &= ~SERIAL_EEPROM_CLK;
			wplxl(Address, Data);
			for ( j = 0 ; j < PLL_DELAY ; j++ ) ; // delay
			// PLL CE = 0
			Data &= ~GP_OUTPUT;
			wplxl(Address, Data);
			for ( j = 0 ; j < PLL_DELAY ; j++ ) ; // delay
			
			for ( i = 0 , Mask = 1 ; i < PLL_MAX_DATA ; i++ , Mask <<= 1 ) {
				// Put Data
				if ( PllData & Mask )
					Data |= SERIAL_EEPROM_DATA;
				else
					Data &= ~SERIAL_EEPROM_DATA;
				wplxl(Address, Data);
				for ( j = 0 ; j < PLL_DELAY ; j++ ) ; // delay
				// PLL CLK = 1
				Data |= SERIAL_EEPROM_CLK;
				wplxl(Address, Data);
				for ( j = 0 ; j < PLL_DELAY ; j++ ) ; // delay
				// PLL CLK = 0
				Data &= ~SERIAL_EEPROM_CLK;
				wplxl(Address, Data);
			}
			for ( j = 0 ; j < PLL_DELAY ; j++ ) ; // delay
			// PLL CLK = 1
			Data |= SERIAL_EEPROM_CLK;
			wplxl(Address, Data);
			// PLL CE = 1
			Data |= GP_OUTPUT;
			wplxl(Address, Data);
			for ( j = 0 ; j < PLL_DELAY ; j++ ) ; // delay
			// PLL CLK = 0
			Data &= ~SERIAL_EEPROM_CLK;
			wplxl(Address, Data);
			// Enable CS of the EEPROM
			Data |= SERIAL_EEPROM_CS;
			wplxl(Address, Data);
		}
		break;
	case KFIR_BOARD_5E:
		switch ( AudioFreq ) {
		case 0: // return "5.5125KHz";
		case 1: // return "8KHz";
		case 2: // return "11.025KHz";
		case 3: // return "16KHz";
		case 4: // return "22.5KHz";
		case 5: // return "32KHz"; 
			Data = 0;
			break;
		case 6: // return "44.1KHz";
			Data = 0x5;
			break;
		default:
		case 7: // return "48KHz";
			Data = 0x6;
		}
		waltl( ALT_ADR_ADSP_FREQ , Data );
		
		break;
	default:
		//ASSERT( 0 );
		break;
	}  
	return 0;
}
#undef PLL_DELAY


static int ADSP_CodeLoad(struct kfir_dev *kfir)
{
	u16 *CodePtr;
	u16 *Boundary;
	u16 Data, BlockSize;

	u32 microlen;
	u16 *micro;

	microlen=firmread("micro/MPEG1_6.HLD", (char**) &micro);
	if (!microlen)
		return -1;

	CodePtr = micro+5;
	Boundary= CodePtr+micro[1];

	AudioRegWrite(kfir, IAL, 0x5000);
	AudioRegWrite(kfir, IDMA, 0xAA55);

	AudioRegWrite(kfir, IAL, 0x5000);
	Data = AudioRegRead(kfir, IDMA);

	while(CodePtr != Boundary) {
		BlockSize = *(CodePtr++);
		AudioRegWrite(kfir, IAL, *(CodePtr++));
		while(BlockSize--) 
			AudioRegWrite(kfir, IDMA, *(CodePtr++));
	}
	vfree(micro);
	udelay(5);
	
	AudioRegWrite(kfir, IAL, 0x1);
	Data = AudioRegRead(kfir, IDMA);
	
	AudioRegWrite(kfir, IAL, 0x1);
	AudioRegWrite(kfir, IDMA,0xAA55);
	
	AudioRegWrite(kfir, IAL, 0x1);
	Data = AudioRegRead(kfir, IDMA);
  

	AudioRegWrite(kfir, IAL, 0x5000);
	udelay(3);
	AudioRegWrite(kfir, IDMA, 0x5555);
	
	udelay(3);

	AudioRegWrite(kfir, IAL, 0x5000);
	udelay(3);
	if (AudioRegRead(kfir, IDMA) != 0x5555) {
		printk("kfir: adsp loading failed!\n");
		return -1;
	}
	printk("kfir: adsp loading successful\n");
	return 0;
}


static int ADSP_KfInit(struct kfir_dev *kfir)
{
        int eRet;
	int nRetries = 0;
	const WORD* RateArray=ADSP_GetRates(kfir, 
					    kfir->Params.AudioMode, 
					    kfir->Params.AudioFormat);
	DWORD Data;
	
	kfir->m_CmdReg = 0xf;
	if ( !kfir->FlValid ) {
	        Data=raltl(ALT_ADR_ADSP_IRQ);
		Data &= ~( ALT_D_ADSP_IRQ_E | ALT_D_ADSP_IRQ_2);
		waltl( ALT_ADR_ADSP_IRQ , Data );
		Sleep(3);
		Data |= ALT_D_ADSP_IRQ_E | ALT_D_ADSP_IRQ_2;
		waltl(ALT_ADR_ADSP_IRQ , Data );
		
		do {
		        if (!ADSP_CodeLoad(kfir)) {
			        kfir->FlValid = 1;
			        break;         
			}
		} while (nRetries++ <= 8 );
		
		if ( nRetries == 8 ) {
		        kfir->FlValid = 0;
			return -1;
		}
	}
	
	Sleep(10); //SleepEx( 10 , 0 );
	kfir->FlValid = 0;
	eRet = -1;
	nRetries = 0;
	do {
	        WORD wd = 0;
		printk("kfir: ADSP initialization..." );
		
		ADSP_SetupPLL( kfir, kfir->Params.AudioFreq );
		Sleep( 100 );
		
		// Reset ADSP.
		
		// Pulse IRQ E and IRQ 2
		Data=raltl(ALT_ADR_ADSP_IRQ);
		Data &= ~(ALT_D_ADSP_IRQ_E | ALT_D_ADSP_IRQ_2);
		waltl( ALT_ADR_ADSP_IRQ , Data );
		
		Data |= ALT_D_ADSP_IRQ_E | ALT_D_ADSP_IRQ_2;
		Data=waltl( ALT_ADR_ADSP_IRQ , Data );
		
		// Some magic
		AudioRegWrite(kfir, IAL,0);
		AudioRegWrite(kfir, IDMA,0x3C0F);
		AudioRegWrite(kfir, IDMA,0x00FC);
		Sleep(10);   // Critical !!!!
		
		// Initialize ADSP.
		wd|=SendDSPApiCommand(kfir, 
				      DSPAPI_SETSTYLE | 
				      ((kfir->Params.AudioFormat+0x41 ) << 8));
		wd|=SendDSPApiCommand(kfir, DSPAPI_SETSAMPLINGRATE | 
				      ADSP_FreqSample[kfir->Params.AudioFreq]);
		wd|=SendDSPApiCommand(kfir, DSPAPI_SETCRCOPTIONS |
				      ((2-kfir->Params.AudioCRC) << 8));
		wd|=SendDSPApiCommand(kfir, DSPAPI_SETCHANNELS | 
				      (kfir->Params.AudioMode << 8));
		wd|=SendDSPApiCommand(kfir, 
				      DSPAPI_SETRESOLUTION | 
				      ((RateArray[kfir->Params.AudioBitRate]/2) << 8));
		wd|=SendDSPApiCommand(kfir, DSPAPI_QUERYSUPPORTRECORD );
		
		if ( !(wd & 0x8000) ) {
		        kfir->FlValid = 1;
			eRet = 0;
			printk("done\n" );
			break;
		}
		printk("failed\n" );
		Sleep(5);
	} while( nRetries++ < 5 );
	
	return eRet;
}


static inline int ADSP_GetRatesNumber(struct kfir_dev *kfir)
{
        if (!kfir->Params.AudioFormat)
	        return 14;
	else
	        return 10;
}


static bool ADSP_KfStart(struct kfir_dev *kfir)
{
        WORD wd=0;
	
	kfir->m_CmdReg = 0xf;
	wd = SendDSPApiCommand(kfir, DSPAPI_STARTRECORD);

	if (wd & 0x8000)
	        return -1;
	return 0;
}

static bool ADSP_KfStop(struct kfir_dev *kfir)
{
	SendDSPApiCommand(kfir, DSPAPI_STOPRECORD);
	return 0;
}

// This function is taken as is from Gadi's new mux.
static int ADSP_GetSamplingFrequency(struct kfir_dev *kfir)
{

	if (kfir->Params.AudioFreq < 0 || kfir->Params.AudioFreq > 7)
		return 32000;
	switch (ADSP_FreqSample[kfir->Params.AudioFreq]) {
		case 0x000: return 44100;
		case 0x100: return 48000;
		case 0x200: return 32000;
		case 0x300: return 22050;
		case 0x400: return 11025;
		case 0x500: return 5512;
		case 0x600: return 16000;
		default:    return 8000;
	}
}

// This function is taken as is from Gadi's new mux.
// Return Audio Frame size in bytes.
static int ADSP_GetAudioFrameSize(struct kfir_dev *kfir)
{
        int frequency = ADSP_GetSamplingFrequency(kfir);
	int size;
	int bitrate;
	const unsigned short *RateArray;

	RateArray = ADSP_GetRates(kfir, kfir->Params.AudioMode, kfir->Params.AudioFormat);
	bitrate = RateArray[kfir->Params.AudioBitRate];
	if (kfir->Params.AudioFormat == VT_KFIR_LAYER_II)
		size = 144 * bitrate * 1000 / frequency;
	else
		size = 48 * bitrate * 1000 / frequency;

  return size;
}

static int ADSP_GetBitRate(struct kfir_dev *kfir)
{
        const unsigned short *RateArray;
	
	RateArray = ADSP_GetRates(kfir, kfir->Params.AudioMode, kfir->Params.AudioFormat);
	
	return RateArray[kfir->Params.AudioBitRate];
}

static WORD ADSP_GetFreqSample(struct kfir_dev *kfir)
{
        return ADSP_FreqSample[kfir->Params.AudioFreq];
}


/* ----------------------------------------------------------------------- */
/* Kfir functions                                                          */

static int KfirGeneric_KfDownloadMuxParams(struct kfir_dev *kfir)
{
	int eRet = 0, i;
	DWORD Data;
	
	for ( i = 0; i < MAX_MUX_PARAMS; i++ ) {
		Data = (kfir->MuxParams[i] >> 16) & 0xFFFF;
		KfirRegWrite( KFIR_202, Data );
		Data = kfir->MuxParams[i] & 0xFFFF;
		KfirRegWrite( KFIR_204 , Data );
	}
	
	return eRet;
}


u8 PAT[] = {
	0x00, 0x00, /* PAT TID */
	0x00, 0x00, /* PAT PID */
	0x00, 0x14, /* size */
	
	0x00, 0xb0, 0x11, 
	0x00, 0x00, 0xcb, 0x00, 0x00, 0x00, 0x00,
	0xe0, 0x10, 0x00, 0x01, 0xe4, 0x00, 
	0x2a, 0xd6, 0x1a, 0xfb,

	0x00, 0x02, /* PMT TID */
	0x04, 0x00, /* PAT PID */
	0x00, 0x1f, /* size */
	
	0x02, 0xb0, 0x1c,
	0x00, 0x01, 0xcb, 0x00, 0x00, 0xe0, 0xa0, 0xf0, 
	0x05, 0x48, 0x03, 0x01, 0x00, 0x00, 0x02, 0xe0, 
	0xa0, 0xf0, 0x00, 0x03, 0xe0, 0x50, 0xf0, 0x00,
	0xae, 0xea, 0x4e, 0x48,
	0x00,

	0x00, 0x00, 
	0x00, 0x00, 
	0x00, 0x00, 
};


static void
CreateTables(struct kfir_dev *kfir)
{
	int size=70, i;
	
	for ( i = 0  ; i < (size / 2) ; i++ )
		kfir->MuxParams[MUX_AUDIO_TABLES+i]=
			(PAT[2*i]<<8) | (PAT[2*i+1]);

	kfir->MuxParams[MUX_AUDIO_ADR_PAT_TABLE_SIZE] = 26;
	kfir->MuxParams[MUX_AUDIO_ADR_PMT_TABLE_SIZE] = 44;
}

static int Kfir_LoadAudioTable( struct kfir_dev *kfir, const DWORD AudioMemBase )
{
        int eRet = 0, i;
	DWORD Data;
	QWORD Scale;
	BYTE FlPesScramblingCntrl = 0;
	BYTE FlPesPriority = 0;
	BYTE FlDataAlignment = 0;
	BYTE Copyright = 0; //gui
	BYTE FlOriginal = 0; //gui
	BYTE FlDSMTrickMode = 0;
	BYTE FlAdditionalCopy = 0;
	BYTE FlPesCrcFlag = 0;
	BYTE FlPesExtension = 0;
	BYTE FlDiscontinuity = 0;
	BYTE FlRnd = 0;
	BYTE FlEPriority = 0;
	BYTE FlOPCR = 0;
	BYTE FlSplicing = 0;
	BYTE FlPrv = 0;
	BYTE Extension = 0;
	DWORD AudioBound = 1;
	BYTE FlFIxed = 1;
	BYTE FlCsps = 0;
	BYTE FlSystemAudioLock = 1;
	BYTE FlSystemVideoLock = 1;
	DWORD VideoBound = 1;
	BYTE FlPacketrateRestriction = 1;
	DWORD AudioBoundScale = 0;
	DWORD AudioBufferSize = 0x50;
	DWORD VideoBoundScale = 1;
	DWORD VideoBufferSize = 0x50;
	DWORD StreamType;

	const WORD* RateArray=ADSP_GetRates(kfir, kfir->Params.AudioMode, 
					    kfir->Params.AudioFormat);
	
	for ( i = 0 ; i < MAX_MUX_PARAMS ; i++ ) {
		kfir->MuxParams[ i ] = 0;
	}

	switch( kfir->Params.StreamType ) {
	case VT_KFIR_AUDIO_ES:
		StreamType = KFIR_D_AUDIO_ES;
		break;
	case VT_KFIR_SYSTEM:
		StreamType = KFIR_D_SYSTEM;
		break;
	case VT_KFIR_VIDEO_ES:
	case VT_KFIR_A_V_ES:
		StreamType = KFIR_D_VEDEO_ES;
		break;
	case VT_KFIR_PROGRAM:
		StreamType = KFIR_D_PROGRAM;
		break;  
	case VT_KFIR_TRANSPORT:
		StreamType = KFIR_D_TRANSPORT;
		break;
	case VT_KFIR_PACKET_VES:
		StreamType = KFIR_D_PACKET_VIDEO_ES;
		break;
	case VT_KFIR_PACKET_AES:
		StreamType = KFIR_D_PACKET_AUDIO_ES;
		break;
	default:
		StreamType = KFIR_D_PROGRAM;
		break;
	}

	kfir->MuxParams[MUX_AUDIO_ADR_STREAM_TYPE] = StreamType;
	kfir->MuxParams[MUX_AUDIO_ADR_PROGRAM_NUMBER] = kfir->Params.ProgramNumber;  
	kfir->MuxParams[MUX_AUDIO_ADR_TRANSPORT_STREAM_ID] = kfir->Params.TransportStreamID;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_STREAM_TYPE] = DEFAULT_AUDIO_STREAM_TYPE;
	kfir->MuxParams[MUX_AUDIO_ADR_PMT_PID] = kfir->Params.ProgramMapID & _13_BIT_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_VIDEO_PID] = kfir->Params.VideoProgramID & _13_BIT_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_PID] = kfir->Params.AudioProgramID & _13_BIT_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_VIDEO_STREAM_ID] = kfir->Params.VideoStreamID & _8_BIT_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_STREAM_ID] = kfir->Params.AudioStreamID & _8_BIT_MASK;
	Data = kfir->Params.BitRate;  // in bits !!!
	kfir->MuxParams[MUX_AUDIO_ADR_VIDEO_BITRATE_LOW] = Data & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_VIDEO_BITRATE_HIGH] = (Data>>16) & KFIR_GENERAL_MASK;
	
	Data = RateArray[kfir->Params.AudioBitRate] * 1000;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_BITRATE_LOW] = Data & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_BITRATE_HIGH] = (Data>>16) & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_TRANSPORT_BITRATE_LOW] = kfir->Params.TransportRate & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_TRANSPORT_BITRATE_HIGH] = (kfir->Params.TransportRate >> 16) & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_VPTS_OFFSET_LOW] = kfir->Params.InitialVideoPTS & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_VPTS_OFFSET_HIGH] = (kfir->Params.InitialVideoPTS >> 16) & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_APTS_OFFSET_LOW] = kfir->Params.InitialAudioPTS & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_APTS_OFFSET_HIGH] = (kfir->Params.InitialAudioPTS >> 16) & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_MAX_VIDEO_PES_SIZE] = DEFAULT_VIDEO_PES_SIZE; 
	kfir->MuxParams[MUX_AUDIO_ADR_MAX_AUDIO_PES_SIZE] = DEFAULT_AUDIO_PES_SIZE; 
	kfir->MuxParams[MUX_AUDIO_ADR_PAT_RATE] = DEFAULT_PAT_RATE; 
	kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC] =
		kfir->Params.Video_Format == VT_KFIR_NTSC ? 2997 : 2500;

	// Additional parameters for 3/2 pulldown
	kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC_SEC] = 2400; 
	
	Scale = div64(kfir->Params.TransportRate * 100,
		      kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC], 1<<13);

	kfir->MuxParams[MUX_AUDIO_ADR_PCR_BITLIMIT_LOW] = (DWORD)(Scale & KFIR_GENERAL_MASK);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_BITLIMIT_MID] = (DWORD)((Scale >> 16) & KFIR_GENERAL_MASK);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_BITLIMIT_HIGH] = (DWORD)((Scale >> 32) & KFIR_GENERAL_MASK);
	
	// Additional parameters for 3/2 pulldown
	Scale = div64(kfir->Params.TransportRate * 100,
		      kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC_SEC], 1<<13);

	kfir->MuxParams[MUX_AUDIO_ADR_PCR_BITLIMIT_LOW_SEC] = (DWORD)(Scale & KFIR_GENERAL_MASK);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_BITLIMIT_MID_SEC] = (DWORD)((Scale >> 16) & KFIR_GENERAL_MASK);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_BITLIMIT_HIGH_SEC] = (DWORD)((Scale >> 32) & KFIR_GENERAL_MASK);
	
	Scale=div64(_27_MEGA * 100,    // MUX_STEP_SHIFT
		    kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC], 1<<16);

	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_LOW] = (DWORD)(Scale & KFIR_GENERAL_MASK);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_MID] = (DWORD)((Scale >> 16) & KFIR_GENERAL_MASK);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_HIGH] = (DWORD)((Scale >> 32) & KFIR_GENERAL_MASK); 
	
	// Additional parameters for 3/2 pulldown
	Scale=div64(_27_MEGA * 100,    // MUX_STEP_SHIFT
		    kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC_SEC], 1<<16);

	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_LOW_SEC] = (DWORD)(Scale & KFIR_GENERAL_MASK);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_MID_SEC] = (DWORD)((Scale >> 16) & KFIR_GENERAL_MASK);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_HIGH_SEC] = (DWORD)((Scale >> 32) & KFIR_GENERAL_MASK); 
	
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_100SEC_LOW] = ((QWORD)_27_MEGA * 100L) & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_STEP_100SEC_HIGH] = ((QWORD)_27_MEGA * 100 >> 16) & KFIR_GENERAL_MASK;
	
	Scale=div64(_90_KILO * 100,    // MUX_STEP_SHIFT
		    kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC], 1<<16);

	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_STEP_LOW] = (DWORD)(Scale & KFIR_GENERAL_MASK);
	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_STEP_HIGH] = (DWORD)((Scale >> 16) & KFIR_GENERAL_MASK);
	
	// Additional parameters for 3/2 pulldown
	Scale=div64(_90_KILO * 100,    // MUX_STEP_SHIFT
		    kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC_SEC], 1<<16);

	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_STEP_LOW_SEC] = (DWORD)(Scale & KFIR_GENERAL_MASK);
	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_STEP_HIGH_SEC] = (DWORD)((Scale >> 16) & KFIR_GENERAL_MASK);
	
	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_STEP_100SEC_LOW] = (_90_KILO * 100) & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_STEP_100SEC_HIGH] = ((_90_KILO  * 100) >> 16) & KFIR_GENERAL_MASK;
	
	kfir->MuxParams[MUX_AUDIO_ADR_PPH_C1] = 2 << 6 | FlPesScramblingCntrl << 4 | FlPesPriority << 3 | 
		FlDataAlignment << 2 | Copyright << 1 | FlOriginal;
	kfir->MuxParams[MUX_AUDIO_ADR_PPH_C2] = FlDSMTrickMode << 3 | FlAdditionalCopy << 2 | FlPesCrcFlag << 1 | FlPesExtension;
	kfir->MuxParams[MUX_AUDIO_ADR_TPH_C1] = FlDiscontinuity << 2 | FlRnd << 1 | FlEPriority;
	kfir->MuxParams[MUX_AUDIO_ADR_TPH_C2] = FlOPCR << 3 | FlSplicing << 2 | FlPrv << 1 | Extension;
	
	kfir->MuxParams[MUX_AUDIO_ADR_ES_RATE_FLAG] = 0; //0xBAD;
	kfir->MuxParams[MUX_AUDIO_ADR_ESCR_FLAG] = 0; //0xBAD; 
	kfir->MuxParams[MUX_AUDIO_ADR_ESCR_BASE_LOW] = 0; //0xBAD;
	kfir->MuxParams[MUX_AUDIO_ADR_ESCR_BASE_MID] = 0; //0xBAD;
	kfir->MuxParams[MUX_AUDIO_ADR_ESCR_BASE_HIGH] = 0; //0xBAD;
	kfir->MuxParams[MUX_AUDIO_ADR_ESCR_EXTENTION] = 0; //0xBAD;
	
	kfir->MuxParams[MUX_AUDIO_ADR_SCRAMBLING] = 0;
	kfir->MuxParams[MUX_AUDIO_ADR_ERROR_INDICATOR] = 0;
	kfir->MuxParams[MUX_AUDIO_ADR_PRIORITY] = 0;
	
	kfir->MuxParams[MUX_AUDIO_ADR_DTS_PER_FRAME] = _90_KILO * 100 / kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC];
	// Additional parameters for 3/2 pulldown
	kfir->MuxParams[MUX_AUDIO_ADR_DTS_PER_FRAME_SEC] = _90_KILO * 100 / kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC_SEC];
	
	Scale=div64(_90_KILO * 100,    // MUX_STEP_SHIFT
		    kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC],
		    kfir->Params.KF_GOP_M);

	kfir->MuxParams[MUX_AUDIO_ADR_PTS_DTS_OFFSET] = (DWORD)(Scale & KFIR_GENERAL_MASK);
	// Additional parameters for 3/2 pulldown
	Scale=div64(_90_KILO * 100,    // MUX_STEP_SHIFT
		    kfir->MuxParams[MUX_AUDIO_ADR_FRAMES_PER_100_SEC_SEC],
		    kfir->Params.KF_GOP_M);

	kfir->MuxParams[MUX_AUDIO_ADR_PTS_DTS_OFFSET_SEC] = (DWORD)(Scale & KFIR_GENERAL_MASK);
	
	Scale=div64(_27_MEGA * 32, kfir->Params.TransportRate, 1024);
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_CORRECTION_CONST_LOW]=
		(DWORD)(Scale & _11_BIT_MASK); 
	kfir->MuxParams[MUX_AUDIO_ADR_PCR_CORRECTION_CONST_HIGH] = 
		(DWORD)((Scale >> 11) & _11_BIT_MASK);
	
	Scale=div64(_90_KILO * 32, kfir->Params.TransportRate, 1024);
	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_CORRECTION_CONST_LOW] = (DWORD)(Scale & _11_BIT_MASK);
	kfir->MuxParams[MUX_AUDIO_ADR_PCRBASE_CORRECTION_CONST_HIGH] = (DWORD)((Scale >> 11) & _11_BIT_MASK);
	
	Data = (kfir->Params.BitRate+(RateArray[kfir->Params.AudioBitRate]*1000))/400;
	kfir->MuxParams[MUX_AUDIO_ADR_PROGRAM_MUX_RATE_LOW]=Data & KFIR_GENERAL_MASK; 
	kfir->MuxParams[MUX_AUDIO_ADR_PROGRAM_MUX_RATE_HIGH]=(Data >> 16) & _6_BIT_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_RATE_BOUND_LOW]=Data & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_RATE_BOUND_HIGH]=(Data >> 16) & _6_BIT_MASK;
	
	Data=(kfir->Params.BitRate+(RateArray[kfir->Params.AudioBitRate]*1000))/400;
	kfir->MuxParams[MUX_AUDIO_ADR_ES_RATE_LOW] = Data & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_ES_RATE_HIGH] = (Data >> 16) & KFIR_GENERAL_MASK;
	
	Data = (1<<24) | (AudioBound<<18) | (FlFIxed<<17) | (FlCsps<<16) |
		(FlSystemAudioLock<<15) | (FlSystemVideoLock<<14) |
		(1<<13) | (VideoBound<<8) | (FlPacketrateRestriction<<7) | 0x7F;
	kfir->MuxParams[MUX_AUDIO_ADR_SHF_LOW] = Data & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_SHF_HIGH] = (Data >> 16) & _9_BITS_MASK;
	
	Data = (kfir->Params.AudioStreamID << 16) |
		(3 << 14) | (AudioBoundScale << 13) | AudioBufferSize;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_PAR_CONST_LOW] = Data & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_PAR_CONST_HIGH] = 
		(Data >> 16) & KFIR_GENERAL_MASK;
	Data = (kfir->Params.VideoStreamID << 16) | (3 << 14) |
		(VideoBoundScale << 13) | VideoBufferSize;
	kfir->MuxParams[MUX_AUDIO_ADR_VIDEO_PAR_CONST_LOW] = Data & KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_VIDEO_PAR_CONST_HIGH]=(Data>>16)&KFIR_GENERAL_MASK;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_MISS_LATENCY] = 15; // gui
	
	// NEW Parameters
	Data = 90000 * 100;
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_MISS_LATENCY+1]=(Data&KFIR_GENERAL_MASK);
	kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_MISS_LATENCY+2]=((Data>>16)&KFIR_GENERAL_MASK);
	
	
	if ( kfir->revision == KFIR_BOARD_5E )
		// 0 - Audio Data is put from the Host.
		// 1 - Audio Data is transfered from Altera.
		kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_PUSH] = 1;
	else
		kfir->MuxParams[MUX_AUDIO_ADR_AUDIO_PUSH] = 0;

	switch( kfir->Params.AudioFreq ) {
	case VT_KFIR_AUDIO_FREQ_5_5125:
	case VT_KFIR_AUDIO_FREQ_8_0000:
	case VT_KFIR_AUDIO_FREQ_11_025:
	case VT_KFIR_AUDIO_FREQ_16_000:
	case VT_KFIR_AUDIO_FREQ_22_050:
	case VT_KFIR_AUDIO_FREQ_44_100:
	default:
		kfir->MuxParams[MUX_AUDIO_ADR_FRAME_SIZE] = 
			(1440 * RateArray[kfir->Params.AudioBitRate]) / 441;
		kfir->MuxParams[MUX_AUDIO_ADR_FRAME_TIME] = 
			AUDIO_FRAME_TIME_32_000 * 90;
		break;
	case VT_KFIR_AUDIO_FREQ_32_000:
		kfir->MuxParams[MUX_AUDIO_ADR_FRAME_SIZE] = 
			(144 * RateArray[kfir->Params.AudioBitRate]) / 32;
		kfir->MuxParams[MUX_AUDIO_ADR_FRAME_TIME] = 
			AUDIO_FRAME_TIME_32_000 * 90;
		break;
	case VT_KFIR_AUDIO_FREQ_48_000:
		kfir->MuxParams[MUX_AUDIO_ADR_FRAME_SIZE] = 
			(144 * RateArray[kfir->Params.AudioBitRate]) / 48;
		kfir->MuxParams[MUX_AUDIO_ADR_FRAME_TIME] = AUDIO_FRAME_TIME_48_000 * 90;
		break;
	}

	CreateTables(kfir);

#if 0
	for ( i = 0 ; i < MAX_MUX_PARAMS ; i++ ) {
		printk("kfir: kfir->MuxParams[%d (%x)] = %x\n",
		       i, i,  kfir->MuxParams[i] );
	}
#endif
	eRet = KfirGeneric_KfDownloadMuxParams( kfir );
	return eRet;
}


static void ReadKfirStateGeneric(struct kfir_dev *kfir)
{
	  kfir->KfState.OldIntCnt = KfirRegRead( KFIR_ADR_INTERRUPT_CNT);
	  kfir->KfState.IntDMA_Mux = KfirRegRead( KFIR_ADR_INT_DMA_MUX);
	  kfir->KfState.IntVideo = KfirRegRead( KFIR_ADR_INT_VIDEO);
	  kfir->KfState.IntAudio = KfirRegRead( KFIR_ADR_INT_AUDIO);
	  kfir->KfState.IntDMA_VIB = KfirRegRead( KFIR_ADR_INT_DMA_VIB);
	  kfir->KfState.MuxCnt_0 = KfirRegRead( KFIR_ADR_MUX_CNT_0);
	  kfir->KfState.MuxCnt_1 = KfirRegRead( KFIR_ADR_MUX_CNT_1);
	  kfir->KfState.MuxCnt_2 = KfirRegRead( KFIR_ADR_MUX_CNT_2);
	  kfir->TotalLinesAfterMUX = kfir->KfState.MuxCnt_0 + 
		  (kfir->KfState.MuxCnt_1 << 16) + 
		  ((QWORD)(kfir->KfState.MuxCnt_2) << 32);
	  kfir->KfState.GOP_0 = KfirRegRead( KFIR_ADR_GOP_0);
	  kfir->KfState.GOP_1 = KfirRegRead( KFIR_ADR_GOP_1);
	  kfir->KfState.Frame = KfirRegRead( KFIR_ADR_FRAME);
	  
	  //kfir->KfState.SRAM_Reg_0 = KfirRegRead( KFIR_ADR_SRAM_REG_0);
	  //kfir->KfState.SRAM_Reg_1 = KfirRegRead( KFIR_ADR_SRAM_REG_1);
	  //kfir->KfState.PictType = KfirRegRead( KFIR_ADR_PICT_TYPE);
	  //kfir->KfState.FIFO_Reg_0 = KfirRegRead( KFIR_ADR_FIFO_REG_0);
	  //kfir->KfState.FIFO_Reg_1 = KfirRegRead( KFIR_ADR_FIFO_REG_1);
	  
	  kfir->KfState.KfirEReady = KfirRegRead( KFIR_ADR_EREADY);
}

static void Kfir_Start(struct kfir_dev *kfir)
{
	DWORD Data;
	printk("eready=%08x\n",KfirRegRead( KFIR_ADR_EREADY ));
	Data = KfirRegRead( KFIR_ADR_RECORD );
	KfirRegWrite(KFIR_ADR_RECORD, KFIR_D_RECORD_START );
}

static void Kfir_Stop(struct kfir_dev *kfir)
{
	kfir->FlStopVideo = 1; 
	/*
	  KfirRegRead(KFIR_ADR_RECORD);
	  KfirRegWrite(KFIR_ADR_RECORD, KFIR_D_RECORD_STOP);
	  KfirRegRead(KFIR_ADR_RECORD);
	*/
}

static void InitParams(struct kfir_dev *kfir)
{
	VT_KFIR_CParams *P=&kfir->Params;
	DWORD video_offset, audio_offset;
	
	P->AspectRatio			= VT_KFIR_ASPECT_4_3;
	P->IntraDCPrecision		= VT_KFIR_PREC_8_BITS;
	P->ColorPrimaries		= VT_KFIR_COLOR_ITUR_624_4BG;
	P->TransferCharacteristics	= VT_KFIR_TRANS_ITUR_624_4BG;
	P->MatrixCoefficients           = VT_KFIR_MATRIX_FCC;
	P->FramePredictionDCT_I	        = VT_KFIR_FIELD_FRAME;
	P->FramePredictionDCT_B	        = VT_KFIR_FIELD_FRAME;
	P->FramePredictionDCT_P	        = VT_KFIR_FIELD_FRAME;
	P->IntraVLCFormat_I		= VT_KFIR_B15;
	P->IntraVLCFormat_B		= VT_KFIR_B14;
	P->IntraVLCFormat_P		= VT_KFIR_B14;
	P->AlternateScan_I		= VT_KFIR_Zigzag;
	P->AlternateScan_B		= VT_KFIR_Zigzag;
	P->AlternateScan_P		= VT_KFIR_Zigzag;
	P->ProgressiveSequence	        = VT_KFIR_OFF;
	P->ProgressiveFrame		= VT_KFIR_OFF;
	P->SliceBypass		        = VT_KFIR_OFF;
	P->VIB_FilterBypass		= VT_KFIR_ON;
	P->VIB_ProgressiveSource	= VT_KFIR_OFF;
	P->Video_Source		        = VT_KFIR_Composite;
	P->FileMode                     = 0;
	P->Video_Format		        = vidoutmode;
	P->AudioMode		        = VT_KFIR_STEREO;
	P->AudioFormat		        = VT_KFIR_LAYER_II;
	P->AudioFreq                    = sndfreq;
	P->AudioBitRate		        = 2;
	P->AudioCRC			= VT_KFIR_OFF;
	P->StreamType	                = streamtype;

	P->KF_MaxBitRate = 9000000;
	P->KF_MinBitRate = 4000000;

	P->TopField	 = 1;
	P->Bottom_Field	 = 1;
	P->Horiz	 = 0;
	P->VBFBufferSize = 112;
	P->BitRate	 = vidrate;
	P->TransportRate = vidrate+300000;// FIXME: calc soundrate and add
	P->EachChunkPTS = VT_KFIR_OFF;
	P->VideoBufferSize=3;
	P->AudioBufferSize=3;
	P->VideoEventSize = 0;
	P->AudioEventSize = 1;

	P->VAD_LuminanceFactor      = 0;	/* 1 */
	P->VAD_LuminanceBandPass    = 1;	/* 1 */		/* use 2!!! */
	P->VAD_Prefilter            = 0;        /* use 1 if noise level is low - cause it's sharping !!! */
	P->VAD_LuminanceBrightness  = 0x80;
	P->VAD_LuminanceContrast    = 0x47;
	P->VAD_ChromaSaturation     = 64;
	P->VAD_HueControl           = 0x0;
	P->VAD_ChromLowPass         = 1;
	P->VAD_ChromCombFilterOff   = 1;
	P->VAD_ColorStandard        = 0;
	
	P->VAD_Dithering=0;
	P->VAD_ChromInterFilter=0;
	
	
	P->AudioStreamID		= 0xC0;
	P->VideoStreamID		= 0xE0;
	P->AudioProgramID		= 0x50;
	P->VideoProgramID		= 0xa0;
	P->ProgramMapID			= 0x400;
	P->SeqStart=1;
	P->FromFrame=0;
	P->ToFrame=100;
	P->PATVersion = 1;
	video_offset = (DWORD) (90000000UL/
				(P->Video_Format == VT_KFIR_PAL)?(2500):(2997));
	audio_offset = (DWORD) (90000000UL/
				(P->Video_Format == VT_KFIR_PAL)?(2500):(2997));
	P->InitialVideoPTS = video_offset;
	P->InitialVideoDTS = video_offset;
	P->InitialAudioPTS = audio_offset;
	P->InitialAudioDTS = audio_offset;
	P->ProgramNumber = 1;
	P->TransportStreamID = 0;
	P->EncodingType = VT_KFIR_LIVEMODE;
	
	// Kfir parameters initialization.
	if (vidoutmode==VT_KFIR_PAL)
		P->KF_FrameRate = KFIR_25;
	else
		P->KF_FrameRate = KFIR_30;
	
	P->KF_HorResolution = (vidsize&7);
	P->KF_VertResolution = ((vidsize>>4)&7);

	P->KF_VESAlignment = KFIR_VES_32;
	P->KF_RateControl = KFIR_RATE_VBR;
	P->KF_QuantScale = 112;
	P->KF_GOP_N = 15;
	P->KF_GOP_M = KFIR_IPBB_PICT;
	P->KF_SIF_PictureBottomField = 0;
	P->KF_LowDelay = 0;
	P->KF_LowSearchRange = 0;
	// P->KF_Auto_3_2_PullDown = 1;
	P->KF_Auto_3_2_PullDown = P->KF_GOP_M == KFIR_IPBB_PICT ? 1 : 0;
	P->KF_Auto_3_2_PullDown = 0;
	P->KF_Threshold = 2;
	
	// Kfir Filtering parameters.
	P->KF_HorLuminFilter = 1;
	P->KF_HorLuminFiltC0 = 38;
	P->KF_HorLuminFiltC1 = -11;
	P->KF_HorLuminFiltC2 = 113;
	P->KF_HorLuminFiltC3 = 924;
	P->KF_HorLuminFiltC4 = 53;
	P->KF_HorLuminFiltC5 = -73;
	P->KF_HorLuminFiltC6 = -22;
	P->KF_HorChromaFilter = 1;
	P->KF_HorChromaFiltC0 = 38;
	P->KF_HorChromaFiltC1 = -11;
	P->KF_HorChromaFiltC2 = 113;
	P->KF_HorChromaFiltC3 = 924;
	P->KF_HorChromaFiltC4 = 53;
	P->KF_HorChromaFiltC5 = -73;
	P->KF_HorChromaFiltC6 = -22;
	P->KF_VertChromaTopFilter = 1;
	P->KF_VertChromaTopFiltCenter = 0;
	P->KF_VertChromaTopFiltC0 = -40;
	P->KF_VertChromaTopFiltC1 = 216;
	P->KF_VertChromaTopFiltC2 = 480;
	P->KF_VertChromaTopFiltC3 = 376;
	P->KF_VertChromaTopFiltC4 = 56;
	P->KF_VertChromaTopFiltC5 = -64;
	P->KF_VertChromaBottomFilter = 1;
	P->KF_VertChromaBottomFiltCenter = 1;
	P->KF_VertChromaBottomFiltC0 = -64;
	P->KF_VertChromaBottomFiltC1 = 56;
	P->KF_VertChromaBottomFiltC2 = 376;
	P->KF_VertChromaBottomFiltC3 = 480;
	P->KF_VertChromaBottomFiltC4 = 216;
	P->KF_VertChromaBottomFiltC5 = -40;
	
	P->KF_Copyright = 0;
	P->KF_Original = 1;
	P->KF_MaxVideoPesSize = 2048;
	P->KF_MaxAudioPesSize = 2048;
	P->KF_AudioMissLatency = 15;
	
	P->KF_TopFiledFirst = 0;  
	P->KF_Convergency = 100;
	P->KF_FrameSkipRatio = 1;
	P->KF_3_2_PullDown = KFIR_AUTO;
	P->DisableWriteFile = 0;
	
	P->KF_Mpeg_I = mpeg1mode;
}


static int Kfir_DownLoadParams(struct kfir_dev *kfir)
{
	double FrameRate;
	double r;
	DWORD mbWidth;
	DWORD mbHeight;
	DWORD Data;
	DWORD MulBitRate = kfir->Params.BitRate * kfir->Params.KF_FrameSkipRatio;
	
	printk("kfir: Download parameters...");

	// 0
	kfir->Par[KFIR_ADR_ASPECT_RATIO / 4] = kfir->Params.AspectRatio + 1;
	KfirRegWrite( KFIR_ADR_ASPECT_RATIO, kfir->Par[KFIR_ADR_ASPECT_RATIO / 4]);
	
	// 1
	kfir->Par[KFIR_ADR_FRAME_RATE_CODE / 4] = kfir->Params.KF_FrameRate + 1;
	KfirRegWrite( KFIR_ADR_FRAME_RATE_CODE, 
		      kfir->Par[KFIR_ADR_FRAME_RATE_CODE / 4]);
	
	// 2 Write(COLOR_PRIM,Board->ptkfir->Params.ColorPrimaries+1);
	kfir->Par[KFIR_ADR_COLOR_PRIM / 4] = kfir->Params.ColorPrimaries + 1;
	KfirRegWrite( KFIR_ADR_COLOR_PRIM, kfir->Par[KFIR_ADR_COLOR_PRIM / 4] );
	
	// 3 Write(TRANSFER_CHAR,Board->ptkfir->Params.TransferCharacteristics+1);
	kfir->Par[KFIR_ADR_TRANSFER_CHAR/4] = kfir->Params.TransferCharacteristics + 1;
	KfirRegWrite( KFIR_ADR_TRANSFER_CHAR, kfir->Par[KFIR_ADR_TRANSFER_CHAR/4] );
	
	// 4 Write(MATRIX_COEFF,Board->ptkfir->Params.MatrixCoefficients+1);
	kfir->Par[KFIR_ADR_MATRIX_COEFF/4] = kfir->Params.MatrixCoefficients + 1;
	KfirRegWrite( KFIR_ADR_MATRIX_COEFF, kfir->Par[KFIR_ADR_MATRIX_COEFF/4] );
	
	// 5 Write(BIT_RATE,(WORD)(Board->ptkfir->Params.BitRate/400));
	Data = MulBitRate/400;
	kfir->Par[KFIR_ADR_BIT_RATE_0/4] = (WORD)(Data & LEAST_BYTE);
	KfirRegWrite( KFIR_ADR_BIT_RATE_0, kfir->Par[KFIR_ADR_BIT_RATE_0/4] );
	
	// 6
	kfir->Par[KFIR_ADR_BIT_RATE_1/4] = (WORD)((Data & SECOND_BYTE) >> 8);
	KfirRegWrite( KFIR_ADR_BIT_RATE_1, kfir->Par[KFIR_ADR_BIT_RATE_1/4] );
	
	// 7
	kfir->Par[KFIR_ADR_BIT_RATE_2/4] = (WORD)((Data & THIRD_BYTE) >> 16);
	KfirRegWrite( KFIR_ADR_BIT_RATE_2, kfir->Par[KFIR_ADR_BIT_RATE_2/4] );
	
	// 8 Write(VBV_BUF_SIZE,Board->ptkfir->Params.VBFBufferSize);
	kfir->Par[KFIR_ADR_VBF_BUFFER_SIZE/4] = kfir->Params.VBFBufferSize;
	KfirRegWrite( KFIR_ADR_VBF_BUFFER_SIZE, kfir->Params.VBFBufferSize );
	
	// 9 Write(PROG_SEQ,Board->ptkfir->Params.ProgressiveSequence);
	//fprintf( ptBoard->VxD->ptFile,  "Progressive sequence:\n" );
	kfir->Par[KFIR_ADR_PROGRESSIVE_SEQ/4] = kfir->Params.ProgressiveSequence;
	KfirRegWrite( KFIR_ADR_PROGRESSIVE_SEQ, kfir->Params.ProgressiveSequence );
	
	// 10 Write(PROG_FRAME,Board->ptkfir->Params.ProgressiveFrame);
	kfir->Par[KFIR_ADR_PROGRESSIVE_FRAME/4] = kfir->Params.ProgressiveFrame;
	KfirRegWrite( KFIR_ADR_PROGRESSIVE_FRAME, kfir->Params.ProgressiveFrame );
	
	// 11 Write(INTRA_DC_PREC,Board->ptkfir->Params.IntraDCPrecision);
	kfir->Par[KFIR_ADR_INTRA_DC_PRECISION/4] = kfir->Params.IntraDCPrecision ;
	KfirRegWrite( KFIR_ADR_INTRA_DC_PRECISION, kfir->Params.IntraDCPrecision );
	
	// 12 Write(INTRA_VLC_FORMAT_I,Board->ptkfir->Params.IntraVLCFormat_I);
	kfir->Par[KFIR_ADR_INTRA_VLC_FORMAT_I/4] = kfir->Params.IntraVLCFormat_I;
	KfirRegWrite( KFIR_ADR_INTRA_VLC_FORMAT_I, kfir->Params.IntraVLCFormat_I );
	
	// 13 Write(INTRA_VLC_FORMAT_P,Board->ptkfir->Params.IntraVLCFormat_P);
	kfir->Par[KFIR_ADR_INTRA_VLC_FORMAT_P/4] = kfir->Params.IntraVLCFormat_P;
	KfirRegWrite( KFIR_ADR_INTRA_VLC_FORMAT_P, kfir->Params.IntraVLCFormat_P );
	
	// 14 Write(INTRA_VLC_FORMAT_P,Board->ptkfir->Params.IntraVLCFormat_B);
	kfir->Par[KFIR_ADR_INTRA_VLC_FORMAT_B/4] = kfir->Params.IntraVLCFormat_B;
	KfirRegWrite( KFIR_ADR_INTRA_VLC_FORMAT_B, kfir->Params.IntraVLCFormat_B );
	
	// 15 Write(INTRA_VLC_FORMAT_I,Board->ptkfir->Params.AlternateScan_I);
	kfir->Par[KFIR_ADR_ALT_SCAN_I/4] = kfir->Params.AlternateScan_I;
	KfirRegWrite( KFIR_ADR_ALT_SCAN_I, kfir->Params.AlternateScan_I );
	
	// 16 Write(INTRA_VLC_FORMAT_P,Board->ptkfir->Params.AlternateScan_P);
	kfir->Par[KFIR_ADR_ALT_SCAN_P/4] = kfir->Params.AlternateScan_P ;
	KfirRegWrite( KFIR_ADR_ALT_SCAN_P, kfir->Params.AlternateScan_P );

	// 17 Write(INTRA_VLC_FORMAT_P,Board->ptkfir->Params.AlternateScan_B);
	kfir->Par[KFIR_ADR_ALT_SCAN_B/4] = kfir->Params.AlternateScan_B;
	KfirRegWrite( KFIR_ADR_ALT_SCAN_B, kfir->Params.AlternateScan_B );
	
	// 18 Write(FRAME_PRED_FRAME_DCT_I,Board->ptkfir->Params.FramePredictionDCT_I);
	kfir->Par[KFIR_ADR_FRAME_PRED_FRAME_DCT_I/4] = 
		kfir->Params.FramePredictionDCT_I;
	KfirRegWrite( KFIR_ADR_FRAME_PRED_FRAME_DCT_I,
		      kfir->Params.FramePredictionDCT_I);
	
	// 19 Write(FRAME_PRED_FRAME_DCT_P,Board->ptkfir->Params.FramePredictionDCT_P);
	kfir->Par[KFIR_ADR_FRAME_PRED_FRAME_DCT_P/4] = 
		kfir->Params.FramePredictionDCT_P;
	KfirRegWrite( KFIR_ADR_FRAME_PRED_FRAME_DCT_P, kfir->Params.FramePredictionDCT_P );
	
	// 20 Write(FRAME_PRED_FRAME_DCT_B,Board->ptkfir->Params.FramePredictionDCT_B);
	kfir->Par[KFIR_ADR_FRAME_PRED_FRAME_DCT_B/4] = kfir->Params.FramePredictionDCT_B;
	KfirRegWrite( KFIR_ADR_FRAME_PRED_FRAME_DCT_B, kfir->Params.FramePredictionDCT_B );
	
	// 21 Write(DSP_PARAM0,AVG_ACT);
	kfir->Par[KFIR_ADR_AVG_ACT/4] = KFIR_D_AVG_ACT;
	KfirRegWrite( KFIR_ADR_AVG_ACT, KFIR_D_AVG_ACT );
	
	// 22 Write(DSP_PARAM1,(WORD)(RMUL(160l,(QWORD)Board->ptkfir->Params.BitRate,115l) >> 17));
	kfir->Par[KFIR_ADR_X_I/4] = (WORD)(RMUL(160l,MulBitRate,115l) >> 17);
	KfirRegWrite( KFIR_ADR_X_I, kfir->Par[KFIR_ADR_X_I/4]);

	// 23 Write(DSP_PARAM2,(WORD)(RMUL(60l,Board->ptkfir->Params.BitRate,115l) >> 17));
	kfir->Par[KFIR_ADR_X_P/4] = (WORD)(RMUL(60l,MulBitRate,115l) >> 17);
	KfirRegWrite( KFIR_ADR_X_P, kfir->Par[KFIR_ADR_X_P/4] );

	// 24 Write(DSP_PARAM3,(WORD)(RMUL(42l,Board->ptkfir->Params.BitRate,115l) >> 17));
	kfir->Par[KFIR_ADR_X_B/4] = (WORD)(RMUL(42l,MulBitRate,115l) >> 17);
	KfirRegWrite( KFIR_ADR_X_B, kfir->Par[KFIR_ADR_X_B/4] );
	
	
	switch ( kfir->Params.KF_HorResolution ) {
	case KFIR_H_160:
		mbWidth = 160 / 16;
		break;
	case KFIR_H_320:
		mbWidth = 320 / 16;
		break;
	case KFIR_H_352:
		mbWidth = 352 / 16;
		break;
	case KFIR_H_480:
		mbWidth = 480 / 16;
		break;
	case KFIR_H_544:
		mbWidth = 544 / 16;
		break;
	case KFIR_H_640:
		mbWidth = 640 / 16;
		break;
	default:
	case KFIR_H_FD1:
		mbWidth = 704 / 16;
		break;
	}

	if (kfir->Params.Video_Format) {  // PAL
		switch ( kfir->Params.KF_VertResolution ) {
		case KFIR_V_144:
			mbHeight = 144 / 16;
			break;
		case KFIR_V_288:
			mbHeight = 288 / 16;
			break;
		default:
		case KFIR_V_576:
			mbHeight = 576 / 16;
			break;    
		}
	}
	else {  // NTSC
		switch ( kfir->Params.KF_VertResolution ) {
		case KFIR_V_112:
			mbHeight = 112 / 16;
			break;
		case KFIR_V_240:
			mbHeight = 240 / 16;
			break;
		default:
		case KFIR_V_480:
			mbHeight = 480 / 16;
			break;
		}
	}
	switch (kfir->Params.KF_FrameRate) {
	case KFIR_23_976:
		FrameRate = 23.976;
		break;
	case KFIR_24:
		FrameRate = 24.0;
		break;
	default:
	case KFIR_25:
		FrameRate = 25.0;
		break;
	case KFIR_29_97:
		FrameRate = 29.97;
		break;
	case KFIR_30:
		FrameRate = 30.0;
		break;
	}
	//DWORD r=((kfir->Params.BitRate << 2)+ FrameRate)/(FrameRate << 1);
	r = (2 * MulBitRate / FrameRate) + 0.5;
	
	// 25 Write(DSP_PARAM4,(WORD)((RMUL(10l,r,31l) >> 2) & MS_LOW));
	kfir->Par[KFIR_ADR_D0_I_0/4] = (WORD)((RMUL(10l,r,31l) >> 2) & KFIR_GENERAL_MASK);
	KfirRegWrite( KFIR_ADR_D0_I_0, (WORD)((RMUL(10l,r,31l) >> 2) & KFIR_GENERAL_MASK));
	
	// 26 Write(DSP_PARAM5,(WORD)((RMUL(10l,r,31l) >> 2) & MS_LOW));
	kfir->Par[KFIR_ADR_D0_P_0/4] = (WORD)((RMUL(10l,r,31l) >> 2) & KFIR_GENERAL_MASK);
	KfirRegWrite( KFIR_ADR_D0_P_0, (WORD)((RMUL(10l,r,31l) >> 2) & KFIR_GENERAL_MASK));
	
	// 27 Write(DSP_PARAM6,(WORD)((RMUL(14l,r,31l) >> 2) & MS_LOW));
	kfir->Par[KFIR_ADR_D0_B_0/4] = (WORD)((RMUL(14l,r,31l) >> 2) & KFIR_GENERAL_MASK);
	KfirRegWrite( KFIR_ADR_D0_B_0, (WORD)((RMUL(14l,r,31l) >> 2) & KFIR_GENERAL_MASK) );
	
	// 28 Write(,(WORD)((RMUL(10l,r,31l) >> 18) & MS_LOW));
	kfir->Par[KFIR_ADR_D0_I_1/4] = (WORD)((RMUL(10l,r,31l) >> 18) & KFIR_GENERAL_MASK);
	KfirRegWrite( KFIR_ADR_D0_I_1, (WORD)((RMUL(10l,r,31l) >> 18) & KFIR_GENERAL_MASK) );
	
	// 29 Write(,(WORD)((RMUL(10l,r,31l) >> 18) & MS_LOW));
	kfir->Par[KFIR_ADR_D0_P_1/4] = (WORD)((RMUL(10l,r,31l) >> 18) & KFIR_GENERAL_MASK);
	KfirRegWrite( KFIR_ADR_D0_P_1, (WORD)((RMUL(10l,r,31l) >> 18) & KFIR_GENERAL_MASK) );
	
	// 30 Write(,(WORD)((RMUL(14l,r,31l) >> 18) & MS_LOW));
	kfir->Par[KFIR_ADR_D0_B_1/4] = (WORD)((RMUL(14l,r,31l) >> 18) & KFIR_GENERAL_MASK);
	KfirRegWrite( KFIR_ADR_D0_B_1, (WORD)((RMUL(14l,r,31l) >> 18) & KFIR_GENERAL_MASK) );
	
	// 31 Write(DSP_PARAM7,9*64);
	kfir->Par[KFIR_ADR_INTRAD/4] = 9*64;
	KfirRegWrite( KFIR_ADR_INTRAD, 9*64 );

	// 32 Write(DSP_PARAM8,(WORD)(RMUL(1l,Board->ptParams->BitRate,4*FrameRate) >> 2));
	// kfir->Par[KFIR_ADR_TMIN/4] = (WORD)(RMUL(1l,kfir->Params.BitRate,4*FrameRate) >> 2);
	kfir->Par[KFIR_ADR_TMIN/4] = (WORD)( (MulBitRate / FrameRate) /4 / 4 );
	KfirRegWrite( KFIR_ADR_TMIN, kfir->Par[KFIR_ADR_TMIN/4] );
	
	// 33
	kfir->Par[KFIR_ADR_R2/4] = (WORD)(RMUL(31L, (1<<22), r));
	KfirRegWrite( KFIR_ADR_R2, (WORD)(RMUL(31L, (1<<22), r)) );
	
	// 34 Write(DSP_PARAM9,(WORD)mbWidth);
	kfir->Par[KFIR_ADR_MB_WIDTH/4] = (WORD)mbWidth;
	KfirRegWrite( KFIR_ADR_MB_WIDTH, (WORD)mbWidth );

	// 35 Write(DSP_PARAM10,(WORD)(0x10000/mbWidth));
	kfir->Par[KFIR_ADR_RMBW/4] = (WORD)(0x10000/mbWidth);
	KfirRegWrite( KFIR_ADR_RMBW, (WORD)(0x10000/mbWidth) );
	
	// 36 Write(DSP_PARAM11,(WORD)(0x10000/(mbWidth*mbHeight)));
	kfir->Par[KFIR_ADR_RMB/4] = (WORD)(0x10000/(mbWidth*mbHeight));
	KfirRegWrite( KFIR_ADR_RMB, (WORD)(0x10000/(mbWidth*mbHeight)));
	
	// 37 Write(DSP_PARAM13,(WORD)(90000l/(Board->ptParams->BitRate >> 13)));
	kfir->Par[KFIR_ADR_GETB/4] = (WORD)(90000L/(MulBitRate >> 13));
	KfirRegWrite( KFIR_ADR_GETB, kfir->Par[KFIR_ADR_GETB/4] );
	
	// 38
	kfir->Par[KFIR_ADR_C1/4] = (WORD)(256 * 1.4);
	KfirRegWrite( KFIR_ADR_C1, (WORD)(256 * 1.4) );
	
	// 39
	kfir->Par[KFIR_ADR_C2/4] = (WORD)((1<<16)/1.4);
	KfirRegWrite( KFIR_ADR_C2, (WORD)((1<<16)/1.4) );
	
	// 40 
	kfir->Par[KFIR_ADR_STD_PICT_DEL/4] = (WORD)(90000.0/FrameRate);
	KfirRegWrite( KFIR_ADR_STD_PICT_DEL, (WORD)(90000.0/FrameRate));
	
	// 41 
	kfir->Par[KFIR_ADR_FIRST_PICT_DEL/4] = 
		(WORD) ((((((double)kfir->Params.VBFBufferSize)*16384.0*2.0)/8.0)*90000.0)/((double)MulBitRate));
	KfirRegWrite( KFIR_ADR_FIRST_PICT_DEL, kfir->Par[KFIR_ADR_FIRST_PICT_DEL/4] );

	// 42 
	//kfir->Par[KFIR_ADR_GOPR1_1/4] = (WORD)((RMUL((kfir->Params.KF_GOP_N-kfir->Params.KF_GOP_M+1),
	//  kfir->Params.BitRate, FrameRate )>>2) >> 16);
	kfir->Par[KFIR_ADR_GOPR1_1/4] = 
		(WORD)(((DWORD)((kfir->Params.KF_GOP_N-
				 kfir->Params.KF_GOP_M+1-
				 (kfir->Params.KF_GOP_M==0))*
				MulBitRate/FrameRate/4))>>16);
	
	KfirRegWrite( KFIR_ADR_GOPR1_1, kfir->Par[KFIR_ADR_GOPR1_1/4] );
	
	// 43 
	// kfir->Par[KFIR_ADR_GOPR2_1/4] = (WORD)((RMUL(kfir->Params.KF_GOP_N, kfir->Params.BitRate,FrameRate)>>2) >> 16);
	kfir->Par[KFIR_ADR_GOPR2_1/4] = (WORD)((DWORD)(kfir->Params.KF_GOP_N*MulBitRate/FrameRate/4) >> 16);
	KfirRegWrite( KFIR_ADR_GOPR2_1, kfir->Par[KFIR_ADR_GOPR2_1/4] );
	
	// 44
	kfir->Par[KFIR_ADR_GOPR1_0/4] = (WORD)((DWORD)((kfir->Params.KF_GOP_N-
							kfir->Params.KF_GOP_M+1-
							(kfir->Params.KF_GOP_M==0))*
						       MulBitRate/FrameRate/4));
	KfirRegWrite( KFIR_ADR_GOPR1_0, kfir->Par[KFIR_ADR_GOPR1_0/4] );
	
	// 45
	kfir->Par[KFIR_ADR_GOPR2_0/4] = (WORD)((DWORD)(kfir->Params.KF_GOP_N *
						       MulBitRate / FrameRate / 4));
	KfirRegWrite( KFIR_ADR_GOPR2_0, kfir->Par[KFIR_ADR_GOPR2_0/4] );
	
	// 46 Write(SLICE_CONTROL,Board->ptParams->SliceBypass);
	kfir->Par[KFIR_ADR_SLICE_CONTROL/4] = kfir->Params.SliceBypass;
	KfirRegWrite( KFIR_ADR_SLICE_CONTROL, kfir->Params.SliceBypass );

	// 47 Write( KFIR_ADR_VIDEO_FORMAT,Board->ptParams->Video_Format);
	kfir->Par[KFIR_ADR_VIDEO_FORMAT/4] = kfir->Params.Video_Format;
	KfirRegWrite( KFIR_ADR_VIDEO_FORMAT, kfir->Params.Video_Format );

	// 48 Write(KFIR_ADR_VIDEO_FORMAT,Board->ptParams->Video_Format);
	kfir->Par[KFIR_ADR_INTERLACED_SOURCE/4] = kfir->Params.VIB_ProgressiveSource;
	KfirRegWrite( KFIR_ADR_INTERLACED_SOURCE, kfir->Params.VIB_ProgressiveSource );

	// 49
	if ( kfir->Params.KF_GOP_M < 0 || kfir->Params.KF_GOP_M > 3 )
		return -1;
	kfir->Par[KFIR_ADR_M/4] = (WORD)kfir->Params.KF_GOP_M;
	KfirRegWrite( KFIR_ADR_M, (WORD)kfir->Params.KF_GOP_M  );

	// 50
	if ( kfir->Params.KF_GOP_M == 0 )
		Data = kfir->Params.KF_GOP_N;
	else 
		Data = (kfir->Params.KF_GOP_N / kfir->Params.KF_GOP_M);
	kfir->Par[KFIR_ADR_MULTIPLE/4] = (WORD)Data;
	KfirRegWrite( KFIR_ADR_MULTIPLE, (WORD)Data );
	
	// 51
	kfir->Par[KFIR_ADR_TOTAL_MBS/4] = (WORD)(mbWidth*mbHeight);
	KfirRegWrite( KFIR_ADR_TOTAL_MBS, (WORD)(mbWidth*mbHeight) );
	
	// 52
	kfir->Par[KFIR_ADR_MB_LINES/4] = (WORD)mbHeight;
	KfirRegWrite( KFIR_ADR_MB_LINES, (WORD)mbHeight );

	// 53
	kfir->Par[KFIR_ADR_RANDOM_SCAN/4] = (WORD)kfir->Params.SeqStart;
	KfirRegWrite( KFIR_ADR_RANDOM_SCAN, kfir->Params.SeqStart );

	// 54
	kfir->Par[KFIR_ADR_VBR_ENABLE/4] = (WORD)kfir->Params.KF_RateControl;
	KfirRegWrite( KFIR_ADR_VBR_ENABLE, (WORD)kfir->Params.KF_RateControl );

	// 55
	if ( kfir->Params.Video_Format == VT_KFIR_PAL ) {
		kfir->Par[KFIR_ADR_TOP_FIRST/4] = 0;
	}
	else {
		kfir->Par[KFIR_ADR_TOP_FIRST/4] = (WORD)kfir->Params.KF_TopFiledFirst;
	}
	KfirRegWrite( KFIR_ADR_TOP_FIRST, kfir->Par[KFIR_ADR_TOP_FIRST/4]);
	
	// 56
	kfir->Par[KFIR_ADR_LOW_SEARCH_RANGE/4] = (WORD)kfir->Params.KF_LowSearchRange;
	KfirRegWrite( KFIR_ADR_LOW_SEARCH_RANGE, (WORD)kfir->Params.KF_LowSearchRange );
	
	// 57
	kfir->Par[KFIR_ADR_CONST_MQUANT/4] = (WORD)(kfir->Params.KF_QuantScale & 0x3FF);
	KfirRegWrite( KFIR_ADR_CONST_MQUANT, (WORD)(kfir->Params.KF_QuantScale & 0x3FF) );
	
	// 58
	kfir->Par[KFIR_ADR_LOW_LATENCY/4] = (WORD)kfir->Params.KF_LowDelay;
	KfirRegWrite( KFIR_ADR_LOW_LATENCY, (WORD)kfir->Params.KF_LowDelay );
	
	// 59
	kfir->Par[KFIR_ADR_AUTO_PULLDOWN/4] = (WORD)kfir->Params.KF_Auto_3_2_PullDown;
	KfirRegWrite( KFIR_ADR_AUTO_PULLDOWN, (WORD)kfir->Params.KF_Auto_3_2_PullDown );
	
	// 60
	// kfir->Params.KF_VESAlignment 
	kfir->Par[KFIR_ADR_BITSTREAM_WIDTH/4] = (WORD)KFIR_D_BITSTREAM_WIDTH_32;
	KfirRegWrite( KFIR_ADR_BITSTREAM_WIDTH, (WORD)KFIR_D_BITSTREAM_WIDTH_32 );
	
	// 61
	kfir->Par[KFIR_ADR_VOB_BASE_ADDR/4] = VOB_BASE_ADDR;
	KfirRegWrite( KFIR_ADR_VOB_BASE_ADDR, VOB_BASE_ADDR );
	
	// 62
	kfir->Par[KFIR_ADR_USE_SDRAM_FIFO/4] = KFIR_D_USE_SDRAM_FIFO_ON;
	KfirRegWrite( KFIR_ADR_USE_SDRAM_FIFO, KFIR_D_USE_SDRAM_FIFO_ON );
	
	// 63
	switch ( kfir->Params.KF_HorResolution ) {
	case KFIR_H_160:
		Data = KFIR_D_VIB_DEC_PATTERN_160;
		break;
	case KFIR_H_320:
		Data = KFIR_D_VIB_DEC_PATTERN_320;
		break;
	case KFIR_H_352:
		Data = KFIR_D_VIB_DEC_PATTERN_352;
		break;
	case KFIR_H_480:
		Data = KFIR_D_VIB_DEC_PATTERN_480;
		break;
	case KFIR_H_544:
		Data = KFIR_D_VIB_DEC_PATTERN_544;
		break;
	case KFIR_H_640:
		Data = KFIR_D_VIB_DEC_PATTERN_640;
		break;
	case KFIR_H_FD1:
		Data = KFIR_D_VIB_DEC_PATTERN_FD1;
		break;
	default:
		break;
	}
	kfir->Par[KFIR_ADR_VIB_DEC_PATTERN/4] = (WORD)Data;
	KfirRegWrite( KFIR_ADR_VIB_DEC_PATTERN, (WORD)Data );
	
	// 64
	switch ( kfir->Params.KF_HorResolution ) {
	case KFIR_H_160:
		Data = KFIR_ADR_VIB_DEC_SIZE_160;
		break;
	case KFIR_H_320:
		Data = KFIR_ADR_VIB_DEC_SIZE_320;
		break;
	case KFIR_H_352:
		Data = KFIR_ADR_VIB_DEC_SIZE_352;
		break;
	case KFIR_H_480:
		Data = KFIR_ADR_VIB_DEC_SIZE_480;
		break;
	case KFIR_H_544:
		Data = KFIR_ADR_VIB_DEC_SIZE_544;
		break;
	case KFIR_H_640:
		Data = KFIR_ADR_VIB_DEC_SIZE_640;
		break;
	case KFIR_H_FD1:
		Data = KFIR_ADR_VIB_DEC_SIZE_FD1;
		break;
	default:
		break;
	}
	kfir->Par[KFIR_ADR_VIB_DEC_SIZE/4] = (WORD)Data;
	KfirRegWrite( KFIR_ADR_VIB_DEC_SIZE, (WORD)Data );
	
	// 65
	Data = (kfir->Params.KF_HorLuminFilter            ? 0x0001 : 0) |
		(kfir->Params.KF_HorChromaFilter           ? 0x0002 : 0) |
		(kfir->Params.Video_Source == VT_KFIR_FILE  ? 0x0040 : 0) |
		(kfir->Params.KF_TopFiledFirst             ? 0x0080 : 0) |
		(kfir->Params.Video_Format == VT_KFIR_NTSC  ? 0      : 0x0100 );
	if ( kfir->Params.KF_VertChromaTopFilter )
		Data |= kfir->Params.KF_VertChromaTopFiltCenter ? 0x000C : 0x0008;
	if ( kfir->Params.KF_VertChromaBottomFilter )
		Data |= kfir->Params.KF_VertChromaBottomFiltCenter ? 0x0030 : 0x0020;
	Data |= 0x0200;
	kfir->Par[KFIR_ADR_VIB_VIDEO_PAR/4] = (WORD)Data;
	KfirRegWrite( KFIR_ADR_VIB_VIDEO_PAR, (WORD)Data );
	
	// 66
	Data = (kfir->Params.Horiz & KFIR_D_VIB_HORI_TOP_MASK);
	kfir->Par[KFIR_ADR_VIB_HORI_TOP_BOT/4] = (WORD)( (Data << 4) + Data );
	KfirRegWrite( KFIR_ADR_VIB_HORI_TOP_BOT, kfir->Par[KFIR_ADR_VIB_HORI_TOP_BOT/4] );
  
	// 67
	Data = ((kfir->Params.TopField << 5) & KFIR_D_VIB_VERT_TOP_MASK) |
		(kfir->Params.Bottom_Field & KFIR_D_VIB_VERT_BOT_MASK);
	kfir->Par[KFIR_ADR_VIB_VERT_TOP_BOT/4] = (WORD)Data;
	KfirRegWrite( KFIR_ADR_VIB_VERT_TOP_BOT, kfir->Par[KFIR_ADR_VIB_VERT_TOP_BOT/4] );
	
	// 68
	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_0/4] = kfir->Params.KF_HorLuminFiltC0;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_YH_0, kfir->Params.KF_HorLuminFiltC0 );
	
	// 69
	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_1/4] = kfir->Params.KF_HorLuminFiltC1;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_YH_1, kfir->Params.KF_HorLuminFiltC1 );

	// 70
	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_2/4] = kfir->Params.KF_HorLuminFiltC2;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_YH_2, kfir->Params.KF_HorLuminFiltC2 );
	
	// 71
	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_3/4] = kfir->Params.KF_HorLuminFiltC3;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_YH_3, kfir->Params.KF_HorLuminFiltC3 );
	
	// 72
	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_4/4] = kfir->Params.KF_HorLuminFiltC4;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_YH_4, kfir->Params.KF_HorLuminFiltC4 );
	
	// 73
	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_5/4] = kfir->Params.KF_HorLuminFiltC5;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_YH_5, kfir->Params.KF_HorLuminFiltC5 );

	// 74
	kfir->Par[KFIR_ADR_VIB_FILTERC_YH_6/4] = kfir->Params.KF_HorLuminFiltC6;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_YH_6, kfir->Params.KF_HorLuminFiltC6 );
	
	// 75
	//fprintf( ptBoard->VxD->ptFile,  "Hor chroma coeff (0-6):\n" );
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_0/4] = kfir->Params.KF_HorChromaFiltC0;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CH_0, kfir->Params.KF_HorChromaFiltC0 );
	
	// 76
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_1/4] = kfir->Params.KF_HorChromaFiltC1;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CH_1, kfir->Params.KF_HorChromaFiltC1 );
	
	// 77
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_2/4] = kfir->Params.KF_HorChromaFiltC2;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CH_2, kfir->Params.KF_HorChromaFiltC2 );

	// 78
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_3/4] = kfir->Params.KF_HorChromaFiltC3;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CH_3, kfir->Params.KF_HorChromaFiltC3 );
	
	// 79
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_4/4] = kfir->Params.KF_HorChromaFiltC4;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CH_4, kfir->Params.KF_HorChromaFiltC4 );
	
	// 80
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_5/4] = kfir->Params.KF_HorChromaFiltC5;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CH_5, kfir->Params.KF_HorChromaFiltC5 );

	// 81
	kfir->Par[KFIR_ADR_VIB_FILTERC_CH_6/4] = kfir->Params.KF_HorChromaFiltC6;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CH_6, kfir->Params.KF_HorChromaFiltC6 );

	// 82
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV0_0/4] = kfir->Params.KF_VertChromaTopFiltC0;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CV0_0, kfir->Params.KF_VertChromaTopFiltC0 );
	
	// 83
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV0_1/4] = kfir->Params.KF_VertChromaTopFiltC1;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CV0_1, kfir->Params.KF_VertChromaTopFiltC1 );

	// 84
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV0_2/4] = kfir->Params.KF_VertChromaTopFiltC2;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CV0_2, kfir->Params.KF_VertChromaTopFiltC2 );

	// 85
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV0_3/4] = kfir->Params.KF_VertChromaTopFiltC3;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CV0_3, kfir->Params.KF_VertChromaTopFiltC3);

	  // 86
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV0_4/4] = kfir->Params.KF_VertChromaTopFiltC4;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CV0_4, kfir->Params.KF_VertChromaTopFiltC4 );

	// 87
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV0_5/4] = kfir->Params.KF_VertChromaTopFiltC5;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CV0_5, kfir->Params.KF_VertChromaTopFiltC5 );

	// 88
	//fprintf( ptBoard->VxD->ptFile,  "Vert top field coeff (0-5):\n" );
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV1_0/4] = kfir->Params.KF_VertChromaBottomFiltC0;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CV1_0, kfir->Params.KF_VertChromaBottomFiltC0 );
	
	// 89
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV1_1/4] = kfir->Params.KF_VertChromaBottomFiltC1;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CV1_1, kfir->Params.KF_VertChromaBottomFiltC1 );

	// 90
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV1_2/4] = kfir->Params.KF_VertChromaBottomFiltC2;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CV1_2, kfir->Params.KF_VertChromaBottomFiltC2 );

	// 91
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV1_3/4] = kfir->Params.KF_VertChromaBottomFiltC3;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CV1_3, kfir->Params.KF_VertChromaBottomFiltC3 );

	// 92
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV1_4/4] = kfir->Params.KF_VertChromaBottomFiltC4;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CV1_4, kfir->Params.KF_VertChromaBottomFiltC4 );

	// 93
	kfir->Par[KFIR_ADR_VIB_FILTERC_CV1_5/4] = kfir->Params.KF_VertChromaBottomFiltC5;
	KfirRegWrite( KFIR_ADR_VIB_FILTERC_CV1_5, kfir->Params.KF_VertChromaBottomFiltC5 );
	
	// 94
	Data = ((0x1<<kfir->Params.KF_Threshold) - 1)*mbWidth*mbHeight*256/2;
	kfir->Par[KFIR_ADR_TELE_THRESHOLD_0/4] = (WORD)(Data & KFIR_GENERAL_MASK);
	KfirRegWrite( KFIR_ADR_TELE_THRESHOLD_0, kfir->Par[KFIR_ADR_TELE_THRESHOLD_0/4] );
	
	// 95
	kfir->Par[KFIR_ADR_TELE_THRESHOLD_1/4] = (WORD)((Data >> 16) & KFIR_GENERAL_MASK);
	KfirRegWrite( KFIR_ADR_TELE_THRESHOLD_1, kfir->Par[KFIR_ADR_TELE_THRESHOLD_1/4] );
	
	// 96
	kfir->Par[KFIR_ADR_DMA_BSM_BURST/4] = 200;// 64;
	KfirRegWrite( KFIR_ADR_DMA_BSM_BURST, kfir->Par[KFIR_ADR_DMA_BSM_BURST/4]);
	
	// 97
	kfir->Par[KFIR_ADR_DMA_BSM_BURST_END/4] = 10;// 60;
	KfirRegWrite( KFIR_ADR_DMA_BSM_BURST_END, kfir->Par[KFIR_ADR_DMA_BSM_BURST_END/4] );

	// 98
	kfir->Par[KFIR_ADR_DMA_VIB_BURST/4] = 200;
	KfirRegWrite( KFIR_ADR_DMA_VIB_BURST, kfir->Par[KFIR_ADR_DMA_VIB_BURST/4] );

	// 99
	kfir->Par[KFIR_ADR_DMA_VIB_BURST_END/4] = 10;
	KfirRegWrite( KFIR_ADR_DMA_VIB_BURST_END, kfir->Par[KFIR_ADR_DMA_VIB_BURST_END/4] );

	kfir->Par[KFIR_ADR_FILE_MODE/4] = kfir->Params.Video_Source == VT_KFIR_FILE ? 1 : 0;
	KfirRegWrite( KFIR_ADR_FILE_MODE, kfir->Par[KFIR_ADR_FILE_MODE/4] );
	
	{
		DWORD gopr = (DWORD)(kfir->Params.KF_GOP_N*MulBitRate/FrameRate)/4;
		DWORD vtmp1;
		DWORD init_T;
		
		if(kfir->Params.KF_GOP_M == 0)
			vtmp1 = 1;
		else
			vtmp1 = (5 + 
				 2 *(kfir->Params.KF_GOP_N/kfir->Params.KF_GOP_M - 1) + 
				 (kfir->Params.KF_GOP_N -
				  (kfir->Params.KF_GOP_N/kfir->Params.KF_GOP_M - 1)-1) );


		init_T = (DWORD)(gopr/vtmp1);
		
		kfir->Par[101*4/4] = (WORD)((init_T >> 16) & 0xFFFF);
		KfirRegWrite( 101 * 4, kfir->Par[101*4/4] );
		
		kfir->Par[102*4/4] = (WORD)( init_T & 0xFFFF);
		KfirRegWrite( 102 * 4, kfir->Par[102*4/4] );
	}

	kfir->Par[KFIR_ADR_MPEG_I_NEG/4] = (WORD)(! (kfir->Params.KF_Mpeg_I) );
	KfirRegWrite( KFIR_ADR_MPEG_I_NEG, kfir->Par[KFIR_ADR_MPEG_I_NEG/4] );
	
	kfir->Par[KFIR_ADR_FRAME_SKIP_RATIO/4] = kfir->Params.KF_FrameSkipRatio;
	KfirRegWrite( KFIR_ADR_FRAME_SKIP_RATIO, kfir->Par[KFIR_ADR_FRAME_SKIP_RATIO/4] );

	printk("kfir: done\n");
	return 0;
}

static int Kfir_ReadNCheckParams( struct kfir_dev *kfir)
{
	int eRet = 0;
	int i;
	DWORD Data[128];
	
	for ( i=0; i<128; i++ ) {
		Data[i] = 0xffff & KfirRegRead( i*4 );
		if ( Data[i] != kfir->Par[i] )
			printk("kfir: Error: Load kfir->Par[%d] = %x, Read Data[%d] = %x\n",
			       i, kfir->Par[i], i, Data[i]);
	}
	
  	return eRet;
}

static int KfirGeneric_StartNDownLoadParams(struct kfir_dev *kfir)
{
	int eRet;
	int Step;
	DWORD Data;

	AlteraRegWrite(ALT_ADR_CLOCK_PADI_DEF, 0);
	
	KfirRegWrite(KFIR_200, DEV_ID_GC << 8);
	KfirRegWrite(KFIR_201, 0);

	KfirRegWrite(KFIR_200 , (DEV_ID_GC<<8) + 5);
	Data = KfirRegRead( KFIR_201 );

	Step = 0;
	do {
		Sleep(100);
		
		if ( Step++ == 10 )
			return -1;
		
		// Read From 254
		Data = (0xffff&KfirRegRead( KFIR_ADR_EREADY ));
	} while ( Data != KFIR_D_EREADY_READY );
	
	// Load ptParams.
	if ( (eRet = Kfir_DownLoadParams(kfir))<0) {
		return eRet;
	}
	// Read and check ptParams.
	if ((eRet = Kfir_ReadNCheckParams(kfir))<0) {
		return eRet;
	}
	// Load Audio Table.
	if ( (eRet = Kfir_LoadAudioTable(kfir, KF_AUDIO_MEM_BASE)) <0 ) {
		return eRet;
	}
	return 0;
}


static void Kfir_Reset(struct kfir_dev *kfir)
{
	if ( kfir->kfirreset)
		return;

	printk("kfir: Reset Kfir and PLL ... ");

	AlteraRegWrite(ALT_ADR_CLOCK_PADI_DEF, 0);

	waltl(ALT_ADR_RESET_REG , 0);
	mdelay(10);
	waltl(ALT_ADR_RESET_REG , ALT_D_RESET_REG_KFIR_PLL_REV_5E);
	mdelay(10);
	waltl(ALT_ADR_RESET_REG , 0);
	mdelay(10);
	waltl(ALT_ADR_RESET_REG , ALT_D_RESET_REG_KFIR_REV_5E);
	mdelay(3);

        kfir->kfirreset=1;

	printk("done\n");
}

#define X_MAX   43
#define Y_MAX   30
static int Kfir_CheckDRAM( struct kfir_dev *kfir)
{
	DWORD Value1 , Value2;
	DWORD Data , Data1;
	int x;
	int y;
	
	for ( y = 0 ; y < Y_MAX ; y++ ) {
		for ( x = 0 ; x < X_MAX ; x++ ) {
			KfirRegWrite(KFIR_ADR_TEST_0, 0xFFFFFFFF);
			KfirRegWrite(KFIR_ADR_TEST_1 , 0xFFFFFFFF );
			Value1 = KfirRegRead( KFIR_ADR_TEST_0 );
			Value2 = KfirRegRead( KFIR_ADR_TEST_1 );

			KfirRegWrite( KFIR_ADR_X , x ); 
			KfirRegWrite( KFIR_ADR_Y , y ); 
			KfirRegWrite( KFIR_ADR_TEST_MODE , KFIR_D_TEST_MODE_ON );

			Data = KfirRegRead( KFIR_ADR_EREADY );

			KfirRegWrite(KFIR_ADR_RECORD, KFIR_D_RECORD_START );
			do {
				Data1 = 0xffff&KfirRegRead( KFIR_ADR_EREADY );  //
				Sleep( 1 );
			} while ( Data1 != 1 );

			Value1 = 0xffff&KfirRegRead( KFIR_ADR_TEST_0 );  // 
			Value2 = 0xffff&KfirRegRead( KFIR_ADR_TEST_1 );  // 
			if ( Value1 || Value2 ) {
				printk( "kfir:DRAM Check Failed. x = %d , y = %d, %08x, %08x\n" , x , y, Value1, Value2);
				// break;
			}
		}
	}
  
	KfirRegWrite( KFIR_ADR_TEST_MODE , KFIR_D_TEST_MODE_OFF );

	return ( x == X_MAX && y == Y_MAX ) ? 0 : -1;
}
#undef X_MAX
#undef Y_MAX



/* ----------------------------------------------------------------------- */
/* I2C functions                                                           */

static void kfir_bit_setscl(void *data, int state)
{
	struct kfir_dev *kfir = (struct kfir_dev *) data;

	if (state)
		kfir->i2c |= 0x01;
	else
		kfir->i2c &= ~0x01;
        waltl(ALT_ADR_I2C, kfir->i2c); 
        udelay(150);
}

static void kfir_bit_setsda(void *data, int state)
{
	struct kfir_dev *kfir = (struct kfir_dev *) data;

	if (state)
		kfir->i2c |= 0x02;
	else
		kfir->i2c &= ~0x02;
	waltl(ALT_ADR_I2C, kfir->i2c); 
	udelay(150);
}

static int kfir_bit_getscl(void *data)
{
	struct kfir_dev *kfir = (struct kfir_dev *) data;
	int state;
	
	state = raltl(ALT_ADR_I2C) & 0x01 ? 1 : 0;
	return state;
}

static int kfir_bit_getsda(void *data)
{
	struct kfir_dev *kfir = (struct kfir_dev *) data;
	int state;

	state = raltl(ALT_ADR_I2C) & 0x02 ? 1 : 0;
	return state;
}

static void kfir_inc_use(struct i2c_adapter *adap)
{
	MOD_INC_USE_COUNT;
}

static void kfir_dec_use(struct i2c_adapter *adap)
{
	MOD_DEC_USE_COUNT;
}

static int attach_inform(struct i2c_client *client)
{
        struct kfir_dev *kfir = (struct kfir_dev *) client->adapter->data;
	
	if (client->driver->id==I2C_DRIVERID_SAA7110) {
		if (!kfir->decoder) 
			kfir->decoder=client;
		printk("kfir: saa711x decoder attached\n");
	}
	
	printk("kfir: i2c attach [%s]\n", client->name);
        return 0;
}

static int detach_inform(struct i2c_client *client)
{
        //struct kfir_dev *kfir = (struct kfir_dev *) client->adapter->data;
	
	printk("kfir: i2c detach [%s]\n", client->name);
        return  0;
}

static struct i2c_algo_bit_data i2c_algo_template = {
	NULL, 
	kfir_bit_setsda,
	kfir_bit_setscl,
	kfir_bit_getsda,
	kfir_bit_getscl,
	10, 10, 100,
};

static struct i2c_adapter i2c_adap_template = {
	"kfir",
	I2C_HW_B_BT848,
	NULL,
	NULL,
	kfir_inc_use,
	kfir_dec_use,
	attach_inform,
	detach_inform,
	NULL,
};

static struct i2c_client i2c_client_template = {
        "kfir internal",
        -1,
        0,
        0,
        NULL,
        NULL
};

static int init_i2c(struct kfir_dev *kfir)
{
	/* i2c bit_adapter */
	memcpy(&kfir->i2c_adap, &i2c_adap_template, sizeof(struct i2c_adapter));
	memcpy(&kfir->i2c_algo, &i2c_algo_template, sizeof(struct i2c_algo_bit_data));
	memcpy(&kfir->i2c_client, &i2c_client_template, sizeof(struct i2c_client));

	sprintf(kfir->i2c_adap.name+strlen(kfir->i2c_adap.name), " #%d", num_kfir-1);
        kfir->i2c_algo.data = kfir;
        kfir->i2c_adap.data = kfir;
        kfir->i2c_adap.algo_data = &kfir->i2c_algo;
        kfir->i2c_client.adapter = &kfir->i2c_adap;

	kfir_bit_setscl(kfir,1);
	kfir_bit_setsda(kfir,1);

	kfir->i2c_ok = i2c_bit_add_bus(&kfir->i2c_adap);
	return kfir->i2c_ok;
}

static void scani2cbus(struct kfir_dev *kfir)
{
        struct i2c_adapter *adap=&kfir->i2c_adap;
	int i;
        struct i2c_msg msgs[1];
	u8 mm1;

	for (i=0; i<0x7f; i++) {
		msgs[0].flags=0;
		msgs[0].addr=i;
		msgs[0].len=1;
		msgs[0].buf=&mm1;
		if (i2c_transfer(adap, msgs, 1)==1)
			printk("kfir: Found i2c device at: 0x%02x\n", 2*i);
	}
}

static int decoder_command(struct kfir_dev *kfir, unsigned int cmd, void *arg)
{
        if (!kfir->decoder)
                return -1;
        return kfir->decoder->driver->command(kfir->decoder, cmd, arg);
}


/****************************************************************************/
/* Altera functions                                                         */

static int AlteraWriteMicro(struct kfir_dev *kfir, u8 *micro, u32 microlen)
{
	u32 nConfig, EEPROMReg, n;
	u32 Data, Data1, GPInput;
  	u8 bVal;
	int i;
  
	// get register with nConfig.
	pci_read_config_dword(kfir->pdev, PLX9054_CONF_HOT_SWAP_OFFSET, &nConfig);
	// get register with DCLK and Data.
	EEPROMReg=rplxl(PLX9054_CNTRL_OFFSET);

	// Set nConfig = 0;
	Data = nConfig & ~HOT_SWAP_LED;
	pci_write_config_dword(kfir->pdev, PLX9054_CONF_HOT_SWAP_OFFSET, Data);
	// Set DCLK = 0;
	Data1 = EEPROMReg & ~SERIAL_EEPROM_CLK;
	wplxl(PLX9054_CNTRL_OFFSET, Data1);
	
	udelay(2);
	
	// Set nConfig = 1
	Data = nConfig | HOT_SWAP_LED;
	pci_write_config_dword(kfir->pdev, PLX9054_CONF_HOT_SWAP_OFFSET, Data);
	
	udelay(2);
	
	GPInput=rplxl(PLX9054_CNTRL_OFFSET);
	
	// begin download code loop
	n = 0;			// initial offset
	Data1 = EEPROMReg;
	while( n < microlen) {
		bVal = micro[n++];
		for ( i = 0 ; i < 8 ; i++ ) {
			// set DCLK = 0
			Data1 = Data1 & ~SERIAL_EEPROM_CLK;
			wplxl(PLX9054_CNTRL_OFFSET, Data1);
			udelay(1);
			
			// Set Data bit.
			if (bVal & 0x01)
				Data1 |= SERIAL_EEPROM_DATA;	// data high
			else
				Data1 &= ~SERIAL_EEPROM_DATA;  //  data Low
			// Output Data with clock low
			wplxl(PLX9054_CNTRL_OFFSET, Data1);
			udelay(1);

			// set DCLK = 1
			Data1 |= SERIAL_EEPROM_CLK;
			wplxl(PLX9054_CNTRL_OFFSET, Data1);
			udelay(1);
			
			bVal >>= 1;
		}
	}
	Data1 &= ~SERIAL_EEPROM_CLK;
	wplxl(PLX9054_CNTRL_OFFSET, Data1);
  
	udelay(2);
	
	// begin Altera init loop
	for ( i = 0 ; i < 20 ; i++ ) {
		// Set DCLK = 0;
		Data1 &= ~SERIAL_EEPROM_CLK;
		wplxl(PLX9054_CNTRL_OFFSET, Data1);
		
		udelay(1);
		
		// set DCLK = 1
		Data1 |= SERIAL_EEPROM_CLK;
		wplxl(PLX9054_CNTRL_OFFSET, Data1);
		
		udelay(1);
	}
	// Set DCLK = 0;
	Data1 = EEPROMReg & ~SERIAL_EEPROM_CLK;
	Data1 &= ~SERIAL_EEPROM_DATA;
	wplxl(PLX9054_CNTRL_OFFSET, Data1);

	GPInput=rplxl(PLX9054_CNTRL_OFFSET);

	return 0;
}

static int AlteraLoadMicro(struct kfir_dev *kfir)
{
	u32 LAS0RR_0, LAS0BA_0, MARBR_0, LBRD0_0, DMRR_0, DMLBAM_0;
	u32 DMLBA1_0, DMPBAM_0, LAS1RR_0, LAS1BA_0, LBRD1_0, INTCSR_0, LROMCBLE_0;

	u32 LAS0RR_1, LAS0BA_1, MARBR_1, LBRD0_1, DMRR_1, DMLBAM_1;
	u32 DMLBA1_1, DMPBAM_1, LAS1RR_1, LAS1BA_1, LBRD1_1, INTCSR_1, LROMCBLE_1;
	u32 Data;

	u32 microlen;
	u8 *micro;
	int result, i;

	switch (kfir->revision) {
	case KFIR_BOARD_A:
	case KFIR_BOARD_B:
		microlen=firmread("micro/altera.rbf", (char**) &micro);
		break;
	case KFIR_BOARD_C:
	case KFIR_BOARD_D:
		microlen=firmread("micro/altera_keb_plx_c.rbf", (char**) &micro);
		break;
		
	case KFIR_BOARD_5E:
		microlen=firmread("micro/altera_keb_plx_5e.rbf", (char**) &micro);
		break;
	default:
		return -1;
	}
	if (!microlen)
		return -1;
	
	printk("loaded microcode with length %d\n", microlen);
	
	
	Data=rplxl(PLX9054_CNTRL_OFFSET);
	
	cli();
	result=AlteraWriteMicro(kfir, micro, microlen);
	sti();
	vfree(micro);
	if (result<0)
		return result;

	waltl(ALT_ADR_INT_REG_CLR , 0xFF);

	// Check ConfDone
	Data=rplxl(PLX9054_CNTRL_OFFSET);
	if ( Data & GP_INPUT ) {
		printk("kfir: GP_INPUT on\n");
		// Reset Altera.
		// Remember all parameters
		LAS0RR_0=rplxl(PLX9054_LAS0RR_OFFSET);
		LAS0BA_0=rplxl(PLX9054_LB0_BASE_ADR);
		MARBR_0=rplxl(PLX9054_MARBR_OFFSET);
		LBRD0_0=rplxl(PLX9054_LBRD0_OFFSET);
		DMRR_0=rplxl(PLX9054_DMRR_OFFSET);
		DMLBAM_0=rplxl(PLX9054_DMLBAM_OFFSET);
		DMLBA1_0=rplxl(PLX9054_DMLBA1_OFFSET);
		DMPBAM_0=rplxl(PLX9054_DMPBAM_OFFSET);
		LAS1RR_0=rplxl(PLX9054_LAS1RR_OFFSET);
		LAS1BA_0=rplxl(PLX9054_LAS1BA_OFFSET);
		LBRD1_0=rplxl(PLX9054_LBRD1_OFFSET);
		INTCSR_0=rplxl(PLX9054_INT_CNTRL_STS);
		LROMCBLE_0=rplxl(PLX9054_BIGEND_LMISC_PROTAREA);
		
		Data=rplxl(PLX9054_CNTRL_OFFSET);
		Data |= 0x40000000;
		wplxl(PLX9054_CNTRL_OFFSET, Data);
		Sleep(10);
		Data &= ~0x40000000;
		wplxl(PLX9054_CNTRL_OFFSET, Data);
		
		for ( i = 0 ; i < 10000 ; i++ ) ;
		
		wplxl(PLX9054_LAS0RR_OFFSET, LAS0RR_0);
		wplxl(PLX9054_LB0_BASE_ADR, LAS0BA_0);
		wplxl(PLX9054_MARBR_OFFSET, MARBR_0);
		wplxl(PLX9054_LBRD0_OFFSET, LBRD0_0);
		wplxl(PLX9054_DMRR_OFFSET, DMRR_0);
		wplxl(PLX9054_DMLBAM_OFFSET, DMLBAM_0);
		wplxl(PLX9054_DMLBA1_OFFSET, DMLBA1_0);
		wplxl(PLX9054_DMPBAM_OFFSET, DMPBAM_0);
		wplxl(PLX9054_LAS1RR_OFFSET, LAS1RR_0);
		wplxl(PLX9054_LAS1BA_OFFSET, LAS1BA_0);
		wplxl(PLX9054_LBRD1_OFFSET, LBRD1_0);
		wplxl(PLX9054_INT_CNTRL_STS, INTCSR_0);
		wplxl(PLX9054_BIGEND_LMISC_PROTAREA, LROMCBLE_0);
		
		// DEBUG
		LAS0RR_1=rplxl(PLX9054_LAS0RR_OFFSET);
		LAS0BA_1=rplxl(PLX9054_LB0_BASE_ADR);
		MARBR_1=rplxl(PLX9054_MARBR_OFFSET);
		LBRD0_1=rplxl(PLX9054_LBRD0_OFFSET);
		DMRR_1=rplxl(PLX9054_DMRR_OFFSET);
		DMLBAM_1=rplxl(PLX9054_DMLBAM_OFFSET);
		DMLBA1_1=rplxl(PLX9054_DMLBA1_OFFSET);
		DMPBAM_1=rplxl(PLX9054_DMPBAM_OFFSET);
		LAS1RR_1=rplxl(PLX9054_LAS1RR_OFFSET);
		LAS1BA_1=rplxl(PLX9054_LAS1BA_OFFSET);
		LBRD1_1=rplxl(PLX9054_LBRD1_OFFSET);
		INTCSR_1=rplxl(PLX9054_INT_CNTRL_STS);
		LROMCBLE_1=rplxl(PLX9054_BIGEND_LMISC_PROTAREA);

		// Mask interupts in Altera.
		waltl(ALT_ADR_INT_MASK , 0);
		
		// Clear the interrupts in Altera.
		Data = raltl(ALT_ADR_INT_REG_CLR);
		waltl(ALT_ADR_INT_REG_CLR , 0xFF);
		/*		
		if ( kfir->revision == KFIR_BOARD_C || kfir->revision == KFIR_BOARD_D ) {
			// Reset the board.
			Data = raltl(ALT_ADR_RESET_REG);
			Data &= ~(ALT_D_RESET_REG_ADSP | ALT_D_RESET_REG_ATOD_REV_C 
				  | ALT_D_RESET_REG_GENNUM_REV_C);
			waltl(ALT_ADR_RESET_REG, Data);
			for ( i = 0 ; i < 100 ; i++ );
			Data |= (ALT_D_RESET_REG_ADSP | ALT_D_RESET_REG_ATOD_REV_C 
				 | ALT_D_RESET_REG_GENNUM_REV_C);
			waltl(ALT_ADR_RESET_REG , Data);
		}
		*/
	}

	return result;
}

static void start_dma0(struct kfir_dev *kfir)
{
	DWORD Data;

	// DMA 0 PCI start address
	wplxl(PLX9054_DMA0_PCI_ADDRESS, virt_to_bus(kfir->dmabuf));

	// DMA 0 LB start address
	Data = rplxl(PLX9054_LB0_BASE_ADR);
	Data &= LB_SPACE0_ADDRESS_MASK;
	if ( kfir->revision != KFIR_BOARD_5E ) {
		//Data += ALT_ADR_COMPRESSED_DATA_FIFO ;
	}
	else {
#ifdef ALTERA_BURST
		Data += 0x4000;
#else
		Data += KFIR_203 + ALT_ADR_KFIR_REV_5E;
#endif
	}
	wplxl(PLX9054_DMA0_LB_ADDRESS, Data );

	kfir->LastDMASize = MUX_DMA_LENGTH;
	wplxl(PLX9054_DMA0_SIZE, MUX_DMA_LENGTH);
	wplxl(PLX9054_INT_CNTRL_STS, rplxl(PLX9054_INT_CNTRL_STS)|DMA0_INT_ENABLE);
	wplxl(PLX9054_DMA_CMD_STS, 
	      rplxl(PLX9054_DMA_CMD_STS)|DMA0_ENABLE|DMA0_START);
}

// MUST be called when Kfir is in GENERIC mode
static void 
KfProcessGenericModeInterrupt(struct kfir_dev *kfir)
{
	DWORD Data;
	
	do {
		ReadKfirStateGeneric(kfir);	
		
		if ( kfir->FlStopVideo ) { // Stop the chip.
			KfirRegWrite( KFIR_ADR_RECORD, KFIR_D_RECORD_STOP );
			Data = 0xffff&KfirRegRead(KFIR_ADR_RECORD);
			if ( Data == KFIR_D_RECORD_STOP )
				kfir->FlStopVideo = 0; 
		}
		
		if (kfir->KfState.IntVideo & KDIR_D_INT_VIDEO ) { // Kfir Video int.
			if ( kfir->IntVideoCnt == 0 ) { // Start DMA transfer here.
				AlteraRegWrite(ALT_ADR_CLOCK_PADI_DEF, 1);
				start_dma0(kfir);
			}
			// Clear this bit.
			KfirRegWrite(KFIR_ADR_INT_VIDEO, 0);
			kfir->IntVideoCnt++;
		}
		
		// Kfir Audio int.    
		if ( kfir->KfState.IntAudio & KDIR_D_INT_AUDIO ) {   
			KfirRegWrite(KFIR_ADR_INT_AUDIO, 0);
			kfir->IntAudioCnt++;
		}
		
		// Kfir VIB interrupt
		if ( kfir->KfState.IntDMA_VIB & KDIR_D_INT_DMA_VIB  ) { 
			//KfirRegWrite(KFIR_ADR_INT_DMA_VIB, 0 ); 
			kfir->IntVibCnt++;
		}
		// Get int counter once more.
		kfir->KfState.IntCnt = KfirRegRead(KFIR_ADR_INTERRUPT_CNT);
	} while ( kfir->KfState.OldIntCnt != kfir->KfState.IntCnt );
	
	return;
}

static void kfir_irq(int irq, void *dev_id, struct pt_regs * regs)
{
	struct kfir_dev *kfir= (struct kfir_dev *) dev_id;
	u32 plxstat;
	u32 altstat, altmask;
	u32 val;
	int count=0;

	while (1) {
		plxstat=rplxl(PLX9054_INT_CNTRL_STS);
		wplxl(PLX9054_INT_CNTRL_STS, plxstat);
		//dprintk("kfir: irq_stat=%08x\n", plxstat);

		if (!(plxstat&(LOCAL_INT_ACTIVE|DMA0_INT_ACTIVE|(1<<14))))
			return;

		/* Error on PCI bus ? */
		if (plxstat & (1<<14)) {
			pci_read_config_dword(kfir->pdev, PCI_COMMAND, &val);
			if (val&0xf8000000) 
				pci_write_config_dword(kfir->pdev, PCI_COMMAND, val);
		}
		
		/* Altera Interrupt ? */
		if (plxstat & LOCAL_INT_ACTIVE) {
			//dprintk("kfir: local int\n");
			altstat = raltl(ALT_ADR_INT_REG_CLR) & 0xff;
			       
			// audio interrupt
			if ( altstat & ALT_D_INT_REQ_CLR_ADSP_INT ) {
				printk("kfir: adsp int\n");
				waltl(ALT_ADR_INT_REG_CLR, altstat);
				
				altmask = raltl(ALT_ADR_INT_MASK) & 0xFF;
				// Mask Audio Int.
				altmask &= ~ALT_D_INT_REQ_CLR_ADSP_INT;
				waltl(ALT_ADR_INT_MASK , altmask );
				
				if (kfir->Params.StreamType != VT_KFIR_VIDEO_ES ) { 
					// nothing for 5E version
				}
				KfirRegWrite(KFIR_ADR_AUDIO_INT_ACK, 1);
			}
			
			// video interrupt
			if ( altstat & ALT_D_INT_REQ_CLR_KFIR_INT ) {
				waltl(ALT_ADR_INT_REG_CLR, altstat);
				if ( kfir->mode == KFIR_GENERIC_MODE )
					KfProcessGenericModeInterrupt(kfir);
			}
		}
		
		if ( kfir->mode == KFIR_GENERIC_MODE ) { 
			// DMA0 interrupt?
			if (plxstat&DMA0_INT_ACTIVE) {
				DWORD Data;
				int free;

				// Clear DMA 0 interrupt.
				Data = rplxl(PLX9054_DMA_CMD_STS);
				Data |= DMA0_CLEAR_INT;
				Data &= ~DMA0_ENABLE;     // disable it.
				wplxl(PLX9054_DMA_CMD_STS, Data);
				
				if (kfir->KfState.KfirEReady == KFIR_D_EREADY_SEQ_END) {
				}
				if (kfir->KfState.KfirEReady == KFIR_D_EREADY_READY) {
					printk("kfir ready\n");
					//kfir->recording=0;
				}
				if (kfir->recording==1) {
					free=kfir->readp-kfir->writep;
					if (free<=0) free+=MPGBUFSIZE;
					if (free>MUX_DMA_LENGTH) {
						memcpy(kfir->mpgbuf+kfir->writep, 
						       kfir->dmabuf, MUX_DMA_LENGTH);
						kfir->writep+=MUX_DMA_LENGTH;
						kfir->writep&=(MPGBUFSIZE-1);
						wake_up_interruptible(&kfir->wq);
						
					} // else we have an overflow!!
				}
				if (kfir->recording==2) {
					int i;
					u8 *ptr;
					u32 val;
					
					if ( kfir->Params.StreamType == 
					     VT_KFIR_TRANSPORT ) {
						kfir->recording=0;
						break;
					}

					for (i=0; i<MUX_DMA_LENGTH-3; i++) {
						ptr=kfir->dmabuf+i;
						val = ((ptr[0] << 24) | 
						       (ptr[1] << 16) |
						       (ptr[2] << 8) |
						       ptr[3]);
						
						if (val==0x000001b7) {
							
							dprintk("kfir: end code 1 found at %d\n",i);
							//kfir->recording=0;
							//break;
						}
						if (val==0x000001b9) {
							dprintk("kfir: end code 2 found at %d\n",i);
							kfir->recording=0;
							break;
						}

					}

				}
				if (kfir->recording) 
					start_dma0(kfir);

			}

		}
		if ((++count)>10) {
			printk("kfir: irq loop\n");
			wplxl(PLX9054_INT_CNTRL_STS, 0);
		}
	}
}


static int init_plx9054(struct kfir_dev *kfir)
{
	u32 val;
	int result=0;

	// Disable interrupts in PLX board (Global PCI Int , 
	// Global Local Int output , DMA0, DMA1 , Doorbell to PCI, Doorbell to Local)
	val=rplxl(PLX9054_INT_CNTRL_STS);
	val &= ~(PCI_INT_ENABLE | LOCAL_INT_OUTPUT_ENABLE | 
		 DMA0_INT_ENABLE | DMA1_INT_ENABLE );
	// Disable PCI abort interrupt.
	val &= ~PCI_ABORT_INT_ENABLE;
	// Disable Master Retry indefinetely. // Obsolete
	val |= RETRY_ABORT_ENABLE;
	// Enable Local Int input
	val |= LOCAL_INT_INPUT_ENABLE;
	// Disable LSERR interrupts.
	val &= ~(LSERR_ENABLE);
	wplxl(PLX9054_INT_CNTRL_STS , val);
	
	// Enable DMA Ch0 Done and route DMA0 interrupt to PCI,
	// enable fast termination, hold Local Bus Address, enable demand mode.
	val = rplxl(PLX9054_DMA0_MODE);
	val |= (DMA0_DONE_ENABLE | DMA0_INT_TO_PCI);
#ifdef ALTERA_BURST
	val &= ~DMA0_HOLD_LOCAL_ADRESS;
#else
	val |= DMA0_HOLD_LOCAL_ADRESS;
#endif
	if (kfir->subsystem_device == KFIR_GENERIC_MODE_ID)
		val |= DMA0_DEMAND_MODE;
	else
		val &= ~DMA0_DEMAND_MODE;
	// Must be Slow Termination for the revision 5E. This fixed the problem with
	// stucking during DMA transfer for the 16 bit width local bus. 28.11.1999 mgv
	val &= ~DMA0_FAST_TERMINATION;    // Slow termination
	// val |= DMA0_FAST_TERMINATION;    // Fast termination
	// Disable Local Burst and Bterm
	// val &= ~DMA0_BTERM_ENABLE;   // BTERM disable
	val |= DMA0_BTERM_ENABLE;       // BTERM enable
	// val &= ~DMA0_LOCAL_BURST_ENABLE;  // burst disable
#if PLX_BURST
	val |= DMA0_LOCAL_BURST_ENABLE;      // burst enable
#else
	val &= ~DMA0_LOCAL_BURST_ENABLE;
#endif

	if ( kfir->revision == KFIR_BOARD_5E ) {
		val &= ~DMA0_LB_WIDTH_MASK;
		val |= DMA0_LB_WIDTH_16;
	}
	wplxl(PLX9054_DMA0_MODE , val);
	
	// Set direction Local->PCI for the DMA channel 0.
	val = rplxl(PLX9054_DMA0_DESCR_PTR);
	val |= DMA0_LOCAL_TO_PCI_TRANSFER;
	// Disable Scater/Gather interrupt from each descriptor terminate counter.
	val &= ~DMA0_DESCRIPTOR_INT_ENABLE;
	wplxl(PLX9054_DMA0_DESCR_PTR, val);

	// Enable DMA Ch1 Done and route it to PCI.
	val = rplxl(PLX9054_DMA1_MODE);
	val |= ( DMA1_DONE_ENABLE | DMA1_INT_TO_PCI );
	wplxl(PLX9054_DMA1_MODE, val);

	result = request_irq(kfir->pdev->irq, kfir_irq,
                             SA_SHIRQ | SA_INTERRUPT, "kfir", (void *) kfir);
	if (result<0)
		return result;

	// KfInitPlx
	
	
	// Local-to-PCI window (for MUX and VIB).
	// Write Range value.
	//val = (~(ptKfirExt->PCI_Length_LB - 1)) & LB_PCI_RANGE_MASK;
	//wplxl(PLX9054_LB_PCI_RANGE, val);
	// Write Local Bus Address.
	//val = (KFIR_MASTER_BASE_LB_ADR & (~(ptKfirExt->PCI_Length_LB - 1))) & LB_BASE_ADR_MASK;
	//wplxl(PLX9054_LB_BASE_ADR, val);
	// Write PCI bus address.
	//val1 = rplxl(PLX9054_PCI_BASE_ADR);
	//val = ptKfirExt->PCI_BaseAdr_LB & PCI_BASE_ADR_MASK;
	//val |= LB_PCI_MEM_ENABLE;
	//val &= ~LB_PCI_IO_ENABLE;
	//wplxl(PLX9054_PCI_BASE_ADR, val);

	// PCI-to-Local Space 0 (for Kfir and Altera).
	val = rplxl(PLX9054_LB0_BASE_ADR);
	wplxl(PLX9054_LB0_BASE_ADR, val);
	val = rplxl(PLX9054_LB0_BASE_ADR);
	
	// Disable prefetch and set maximim delay for the PCI bus.
	// This solved the problem with the write and immidiate read from the ADSP !!! (23.05.99)
	val = rplxl(PLX9054_LBRD0_OFFSET);
	val |= (SPACE0_PREFETCH_DISABLE | PCI_TARGET_RETRY_DELAY_MAX);
	wplxl(PLX9054_LBRD0_OFFSET, val);
	
	// Set read from local bus to be locked operation.
	val = rplxl(PLX9054_MARBR_OFFSET);
	val |= DIRECT_SLAVE_LOCK_ENABLE;
	// Clear the DMA channels priority
	val &= ~PRIOPITY_MASK;
	val |= PRIORITY_DMA_0;    // Set priority to DMA 0.
	wplxl(PLX9054_MARBR_OFFSET, val);
	
	// PCI-to-Local Space 1 (for Audio).
	rplxl(PLX9054_LB1_BASE_ADR);
	val &= ~LB_SPACE1_ADDRESS_MASK;
	val |= (SPACE1_ENABLED | KFIR_AUDIO_INPUT_BASE_LB_ADR);
	wplxl(PLX9054_LB1_BASE_ADR, val);
	
	// Swap bytes for the Direct master (MUX from Kfir and VIB to Kfir) and Direct Slave (Audio to Kfir)
	val = rplxl(PLX9054_BIGEND_LMISC_PROTAREA);
	val |= (BIGEND_DIRECT_MASTER | BIGEND_SPACE_1 | BIGEND_DMA0);
	wplxl(PLX9054_BIGEND_LMISC_PROTAREA, val);
	
	return result;
}

static int 
KfirGeneric_ContrDownLoadMicroCode2(struct kfir_dev *kfir, KF_CONTR_TYPE eKfirContr)
{
	u32 *microbuf;
	DWORD Data;
	DWORD MicroCodeLength;
	DWORD DevId;
	WORD  MicroCodeInst;
	int MicWidth;  
	int Fl_OK = 1;
	DWORD i;

	printk("kfir: segment: %d  ", eKfirContr);

	// get Global Controller microcode, it's length and Instruction.
	switch( eKfirContr ) {
	case KF_GC:
		MicroCodeInst = 3;
		MicWidth = 1;
		DevId = KF_DEV_ID_GC;
		break;
	case KF_DSP_RISC_MC:
		MicroCodeInst = 3;
		MicWidth = 1;//0;
		DevId = KF_DEV_ID_DSP;
		break;
	case KF_DSP_RISC_CRAM:
		MicroCodeInst = 5;
		MicWidth = 1;
		DevId = KF_DEV_ID_DSP;
		break;
	case KF_DSP_UNIT_MC:
		MicroCodeInst = 7;
		MicWidth = 1;
		DevId = KF_DEV_ID_DSP;
		break;
	case KF_BSM_MC:
		MicroCodeInst = 3;
		MicWidth = 1;
		DevId = KF_DEV_ID_BSM;
		break;
	case KF_MUX_MC:
		MicroCodeInst = 3;
		MicWidth = 1;
		DevId = KF_DEV_ID_MUX;
		break;
	default:
		return -1;
	}
	MicroCodeLength=kfir->micro[kfir->mp];
	microbuf=kfir->micro+kfir->mp+1;
	kfir->mp+=kfir->micro[kfir->mp]+1;

 	KfirRegWrite(KFIR_DEV_INST, (DevId<<8)+1);
	KfirRegWrite(KFIR_EXEC, 0);
	KfirRegWrite(KFIR_DEV_INST, (DevId<<8)+2);
	Data = KfirRegRead(KFIR_EXEC);
	KfirRegWrite(KFIR_DEV_INST, (DevId<<8) + MicroCodeInst);
	
	for ( i = 0 ; i < MicroCodeLength ; i++ ) {
		if (MicWidth == 0)
			KfirRegWrite(KFIR_201, (microbuf[i] & 0xFFFF));
		else {
			KfirRegWrite(KFIR_202, (microbuf[i] >> 16) & 0xFFFF);
			KfirRegWrite(KFIR_201, microbuf[i] & 0xFFFF);
		}
	}
	KfirRegWrite(KFIR_DEV_INST, (DevId<<8) + 1);

	KfirRegWrite(KFIR_EXEC, 0);
	KfirRegWrite(KFIR_DEV_INST, (DevId<<8) + 2);
	Data=KfirRegRead(KFIR_EXEC);

	KfirRegWrite(KFIR_DEV_INST, (DevId<<8) + MicroCodeInst + 1);
	
	for ( i = 0 ; i < MicroCodeLength ; i++ ) {
		Data = KfirRegRead( KFIR_201 );
		if ( MicWidth == 0 )
			Data &= 0xFFFF;
		else 
			Data = (KfirRegRead(KFIR_202) << 16) | (Data & 0xffff);
		if ( Data != microbuf[i] ) {
			Fl_OK = 0;
			//printk("%08x %08x\n", Data, microbuf[i] );
		}
	}
	
	printk(Fl_OK ? "Done\n" : "Failed\n");
	
	if (!Fl_OK) {
		return -1;
	}
	
	return 0;
}

static int KfirLoadMicro(struct kfir_dev *kfir)
{
	volatile DWORD Data;
	int ret=0, i;

	kfir->micro=vtv;

	if (!kfir->DVB_MODE) {
		if ( kfir->Params.KF_Mpeg_I )
			kfir->micro=vtv1;
		else {
			if ( kfir->Params.StreamType == VT_KFIR_TRANSPORT ) {
				printk("loading TS code\n");
				kfir->micro=vtv2;
			}
		}
	}
	
	if (kfir->DVB_MODE == 1) {
		switch (kfir->Params.StreamType) {
		case VT_KFIR_VIDEO_ES:
		case VT_KFIR_PROGRAM:
			kfir->micro=vtv8;
		case VT_KFIR_TRANSPORT:
			kfir->micro=vtv28;
			break;
		case VT_KFIR_SYSTEM:
			kfir->micro=vtv18;
			break;
		default:
			break;
		}
	}

	KfirRegWrite(KFIR_DEV_INST, (DEV_ID_GC<<8) + 6);
	Data = KfirRegRead( KFIR_EXEC );
	
	KfirRegWrite(KFIR_DEV_INST, (DEV_ID_GC<<8) + 5);
	Data = KfirRegRead( KFIR_EXEC );
	
	kfir->mp=0;
	for (i = KF_GC ; i<=KF_MUX_MC ; i++) {
		if ((ret=KfirGeneric_ContrDownLoadMicroCode2(kfir, i))<0) {
			return ret;
		}
	}

	return ret;
}

int KfirGeneric_Init(struct kfir_dev *kfir)
{
	int eRet = 0;
	DWORD i;

	if ( !kfir->FlFirstTime ) {
		// Reset all Kfir registers
		for ( i = 0 ; i < 256 ; i++ ) 
			KfirRegWrite( i * 4 , 0 );

		if (KfirLoadMicro(kfir)<0) return -1;
	}

	KfirGeneric_StartNDownLoadParams(kfir);

	/*	
	if ( kfir->FlFirstTime == 0 ) {
	        if ( Kfir_CheckDRAM(kfir) < 0 ) {
			return -1;
		}
	}
	*/
	kfir->FlFirstTime = 1;
	kfir->FlValid = 1;
	
	return eRet;
}


static void InitIRQs(struct kfir_dev *kfir)
{
	u32 val;

	/* Disable PLX PCI interrupt */
	val=rplxl(PLX9054_INT_CNTRL_STS);
	val &= ~PCI_INT_ENABLE;
	wplxl(PLX9054_INT_CNTRL_STS, val);

	/* Disable all Altera interrupts */
	val=raltl(ALT_ADR_INT_MASK);
	waltl( ALT_ADR_INT_MASK, val&(~ALT_D_INT_CLEAR_ALL));

	/* Clear Altera interrupts */
	val=raltl(ALT_ADR_INT_REG_CLR);
	waltl(ALT_ADR_INT_REG_CLR, val|ALT_D_INT_CLEAR_ALL);

	/* Enable PLX PCI and local (Altera) interrupt */
	val=rplxl(PLX9054_INT_CNTRL_STS);
	val |= (PCI_INT_ENABLE | LOCAL_INT_INPUT_ENABLE);
	val &= ~RETRY_ABORT_ENABLE;
	wplxl(PLX9054_INT_CNTRL_STS, val);
	
	/* Enable Altera Kfir interrupts */
	waltl( ALT_ADR_INT_MASK, ALT_D_INT_REQ_CLR_KFIR_INT);
}


static int InitBoard(struct kfir_dev *kfir)
{
	u32 val;

	InitParams(kfir);


	/* Disable PLX PCI interrupt */
	val=rplxl(PLX9054_INT_CNTRL_STS);
	val &= ~PCI_INT_ENABLE;
	wplxl(PLX9054_INT_CNTRL_STS, val);

	if (!loaded)
		AlteraLoadMicro(kfir);

	Kfir_Reset(kfir);

	ADSP_KfInit(kfir);
	KfirGeneric_Init(kfir);

	init_i2c(kfir);
	scani2cbus(kfir);

	decoder_command(kfir, DECODER_GET_STATUS, &val);
	printk("kfir: decoder status=%x\n", val);
	val=vidinmode;
	decoder_command(kfir, DECODER_SET_NORM  , &val);

	InitIRQs(kfir);

	return 0;
}


static int kfir_open(struct video_device *dev, int flags)
{
	struct kfir_dev *kfir = (struct kfir_dev *) dev;

	kfir->recording=0;
	MOD_INC_USE_COUNT;
	return 0;   
}


static void kfir_close(struct video_device *dev)
{
	struct kfir_dev *kfir = (struct kfir_dev *) dev;

	kfir->recording=2;
	ADSP_KfStop(kfir);
	Kfir_Stop(kfir);

	{
		int i;
		int count=0;
		u16 stat;
		u32 dreq;
		while (kfir->recording) {
			if ((count++) > 5000)
				break;
			mdelay(1);
		}
		count=0;

		
		while ((stat=0xffff&KfirRegRead(KFIR_ADR_EREADY))!=1) {
			dreq=KfirRegRead(KFIR_205);
			if (!(dreq&2)) 
				for (i=0; i<64; i++)
					KfirRegRead(KFIR_203);
			if ((count++) > 1000)
				break;
			mdelay(1);
		}
		
		printk("eready=%08x\n",KfirRegRead( KFIR_ADR_EREADY ));
	}
	MOD_DEC_USE_COUNT;  
}

static unsigned int kfir_poll(struct video_device *dev, struct file *file,
			      poll_table *wait)
{
        //struct kfir_dev *kfir=(struct kfir_dev *) dev;
        unsigned int mask=0;

        return mask;
}

static int kfir_mmap(struct video_device *dev, const char *adr, unsigned long size)
{
	//struct kfir_dev * kfir = (struct kfir_dev *)dev;

	return 0;
}

static int kfir_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	//struct kfir_dev * kfir = (struct kfir_dev *) dev;
	return 0;
}        


static long kfir_write(struct video_device *dev, const char *buf,
		       unsigned long count, int nonblock)
{
	//struct kfir_dev * kfir = (struct kfir_dev *)dev;

	/* not supported yet (Kfir2A or board with extra decoder chip ) */
        return 0;
}


static long record(struct kfir_dev *kfir, char *buf,
		   unsigned long count, int nonblock)
{
	DECLARE_WAITQUEUE(wait, current);
        unsigned long todo = count;
        int blocksize, split;
        
        if (!kfir->recording)
        {
		dprintk("kfir: start recording\n");
		///InitIRQs(kfir);
		KfirGeneric_StartNDownLoadParams(kfir);
		kfir->readp=kfir->writep=0;
		kfir->recording=1;
		kfir->IntVideoCnt=0;
		waltl(ALT_ADR_CONTROL_1, 2);
		ADSP_KfStart(kfir);
		KfirRegRead( 205*4 );
		Kfir_Start(kfir);
        }
        
        while (todo > 0) {
                if (kfir->readp == kfir->writep) {
                        if (nonblock)
                                return count-todo;
                        add_wait_queue(&kfir->wq, &wait);
                        while (1) {
                                set_current_state(TASK_INTERRUPTIBLE);
                                if (kfir->readp != kfir->writep
                                    || signal_pending(current))
                                        break;
                                schedule();
                        }
                        current->state=TASK_RUNNING;
                        remove_wait_queue(&kfir->wq, &wait);
                        if (kfir->readp==kfir->writep) 
                                return 0;
                }

                split=MPGBUFSIZE;
                blocksize=kfir->writep-kfir->readp;
                if (blocksize < 0) {
                        blocksize+=MPGBUFSIZE;
                        split=MPGBUFSIZE-kfir->readp;
                }
                if (blocksize > todo)
                        blocksize=todo;
                if (split < blocksize) {
                        if (copy_to_user(buf, kfir->mpgbuf+kfir->readp, split))
				return -EFAULT;
                        buf+=split;
                        kfir->readp=0;
                        todo-=split;
                        blocksize-=split;
                }
                if (blocksize) {
                        if (copy_to_user(buf, kfir->mpgbuf+kfir->readp, blocksize))
                                return -EFAULT;
                        kfir->readp = (kfir->readp + blocksize) 
				& (MPGBUFSIZE-1);
                        todo-=blocksize;
                        buf+=blocksize;
                }
        }
	return count;
}


static long kfir_read(struct video_device *dev, char *buf,
		      unsigned long count, int nonblock)
{
	struct kfir_dev * kfir = (struct kfir_dev *) dev;

	return record(kfir, buf, count, nonblock);
}

static int kfir_init_done(struct video_device *dev)
{
	return 0;
}



/* template for video_device-structure */
static struct video_device kfir_template = {
	"Kfir",
        VID_TYPE_MPEG_ENCODER,  
        VID_HARDWARE_SAA7146,
        kfir_open,
        kfir_close,
        kfir_read,
	kfir_write,		
	kfir_poll,
	kfir_ioctl,
	kfir_mmap,
	kfir_init_done,
	NULL,
	0,
	0
};

static int kfir_init_device (struct pci_dev *pdev)
{
	u32 class_rev;
        struct kfir_dev *kfir;

	if (num_kfir>=MAX_NUM_KFIR)
		return -1;

	kfir=&kfirs[num_kfir];
	num_kfir++;
#if LINUX_VERSION_CODE >= 0x020300
	pdev->driver_data=kfir;
#endif
	memset(kfir, 0, sizeof(struct kfir_dev));

	kfir->pdev=pdev;
	pci_read_config_dword(kfir->pdev, PCI_CLASS_REVISION, &class_rev);
	kfir->revision=class_rev&0xff;
	kfir->kfirreset=0;
	
	pci_read_config_word(kfir->pdev, PCI_SUBSYSTEM_ID, &kfir->subsystem_device);

	if ( kfir->subsystem_device == KFIR_GENERIC_MODE_ID )
		kfir->mode = KFIR_GENERIC_MODE;
	else
		kfir->mode = KFIR_MASTER_MODE;
	kfir->ALT_ADR_INT_MASK = ALT_ADR_INT_MASK_REV_5E;
	kfir->ALT_ADR_INT_REG_CLR = ALT_ADR_INT_REG_CLR_REV_5E;
	kfir->ALT_ADR_KFIR_AUDIO_BASE = 0xFFFFFFFF;
	kfir->ALT_ADR_RESET_REG = ALT_ADR_RESET_REG_REV_5E;

	kfir->ALT_ADR_ADSP_ADDR = ALT_ADR_ADSP_ADDR_REV_5E;
	kfir->ALT_ADR_ADSP_DATA = ALT_ADR_ADSP_DATA_REV_5E;
	kfir->ALT_ADR_ADSP_IRQ = ALT_ADR_ADSP_IRQ_REV_5E;
	kfir->ALT_ADR_ADSP_FREQ = ALT_ADR_AUDIO_FREQ_REV_5E;
	kfir->ALT_ADR_I2C = ALT_ADR_I2C_REV_5E;
	kfir->ALT_ADR_CONTROL_0 = 0xFFFFFFFF;
	kfir->ALT_ADR_CONTROL_1 = ALT_ADR_OUTPUT_REV_5E;
	kfir->ALT_ADR_CLOCK_PADI = ALT_ADR_CLOCK_PADI_DEF;

#if LINUX_VERSION_CODE >= 0x020300
	kfir->mem0=ioremap(pdev->resource[0].start&PCI_BASE_ADDRESS_MEM_MASK, 0x100);
	kfir->mem1=(char *)(pdev->resource[1].start&PCI_BASE_ADDRESS_MEM_MASK);
	kfir->mem2=ioremap(pdev->resource[2].start&PCI_BASE_ADDRESS_MEM_MASK, 0x200000);
	kfir->mem3=ioremap(pdev->resource[3].start&PCI_BASE_ADDRESS_MEM_MASK, 0x400);
#else
	kfir->mem0=ioremap(pdev->base_address[0]&PCI_BASE_ADDRESS_MEM_MASK, 0x100);
	kfir->mem1=(char *)(pdev->base_address[1]&PCI_BASE_ADDRESS_MEM_MASK);
	kfir->mem2=ioremap(pdev->base_address[2]&PCI_BASE_ADDRESS_MEM_MASK, 0x200000);
	kfir->mem3=ioremap(pdev->base_address[3]&PCI_BASE_ADDRESS_MEM_MASK, 0x400);
#endif

	kfir->AlteraReg=kfir->mem2;
	kfir->KfirReg=kfir->AlteraReg;
	if (kfir->revision == 0x5e)
		kfir->KfirReg+=ALT_ADR_KFIR_REV_5E;


	printk("kfir: Kfir board revision %02x\n", (int) kfir->revision);

	kfir->dmabuf=kmalloc(MUX_DMA_LENGTH, GFP_KERNEL);
	if (!kfir->dmabuf) 
		return -1;

	kfir->mpgbuf=vmalloc(MPGBUFSIZE);
	if (!kfir->mpgbuf) {
		kfree(kfir->dmabuf);
		return -1;
	}


        /* Copy structure */
        memcpy( &kfir->video, &kfir_template, sizeof(struct video_device));
        
        if( 0 > (video_register_device(&kfir->video, VFL_TYPE_GRABBER))) {
                printk(KERN_ERR "kfir: can't register videodevice\n");
                return -1;
        }
        init_waitqueue_head(&kfir->wq);

	init_plx9054(kfir);

	return InitBoard(kfir);
}


static void kfir_remove_device(struct kfir_dev *kfir)
{
	DWORD val;

        printk("kfir: kfir_remove_one()\n");

	val=raltl(ALT_ADR_INT_MASK);
	waltl(ALT_ADR_INT_MASK, val&(~ALT_D_INT_CLEAR_ALL));

	val=raltl(ALT_ADR_INT_REG_CLR);
	waltl(ALT_ADR_INT_REG_CLR, val|ALT_D_INT_CLEAR_ALL);

	if ( kfir->mode == KFIR_GENERIC_MODE ) { 
		/* val=rplxl(DMA0_DONE_ENABLE); 
		val&=~DMA0_DONE_ENABLE; 
		wplxl(DMA0_DONE_ENABLE, val); */
		
		val=rplxl(PLX9054_INT_CNTRL_STS);
		val |= DMA0_ENABLE;
		wplxl(PLX9054_INT_CNTRL_STS, val);
		val |= DMA0_START;
		wplxl(PLX9054_INT_CNTRL_STS, val);
		val &= ~DMA0_ENABLE;
		wplxl(PLX9054_INT_CNTRL_STS, val);
		val |= DMA0_ABORT;
		wplxl(PLX9054_INT_CNTRL_STS, val);
		val &= ~DMA0_CLEAR_INT;
		wplxl(PLX9054_INT_CNTRL_STS, val);
	}


        if(kfir->video.minor!=-1)
                video_unregister_device(&kfir->video);

	i2c_bit_del_bus(&kfir->i2c_adap);

	iounmap(kfir->mem0);
	iounmap(kfir->mem2);
	iounmap(kfir->mem3);

	free_irq(kfir->pdev->irq, kfir);

	if (kfir->dmabuf)
		kfree(kfir->dmabuf);
	if (kfir->mpgbuf)
		vfree(kfir->mpgbuf);
}


#if LINUX_VERSION_CODE < 0x020300

static int find_kfir(void) 
{
	struct pci_dev *dev = NULL;
	int result = 0;
	
	if (!pci_present()) {
		printk(KERN_ERR "kfir: PCI-BIOS not present or not accessable!\n");
		return -1;
	}
        while ((dev = pci_find_device(PCI_VENDOR_ID_VISIONTECH,
				      PCI_DEVICE_ID_VISIONTECH_KFIR_5E, dev))) 
	{
		if (num_kfir) 
			ddelay(100);
		result = kfir_init_device(dev);

                if (result<0)
                        return result;
	}
	if (num_kfir)
		printk(KERN_INFO "kfir: %d kfir card(s) found.\n", num_kfir);

	return num_kfir;
}

#ifdef MODULE

int init_module(void) 
{
	if ( 0 > find_kfir() )
		return -EIO;
	return 0;
}

void cleanup_module(void) 
{
	int i;
	
	for (i=0; i<num_kfir; i++)
		kfir_remove_device(&kfirs[i]);
}
#endif


#else

static int __devinit kfir_init_one (struct pci_dev *pdev,
		const struct pci_device_id *ent)
{
	return kfir_init_device(pdev);
}

static void __devexit kfir_remove_one (struct pci_dev *pdev)
{
	return kfir_remove_device((struct kfir_dev *) pdev->driver_data);
}

static void kfir_suspend(struct pci_dev *pdev)
{
        printk("kfir: kfir_suspend()\n");
}

static void kfir_resume(struct pci_dev *pdev)
{
        printk("kfir: kfir_resume()\n");
}


static struct pci_device_id kfir_pci_tbl[] __devinitdata = {
	{ PCI_VENDOR_ID_VISIONTECH, PCI_DEVICE_ID_VISIONTECH_KFIR_5E,
	  PCI_ANY_ID, PCI_ANY_ID, },
	{ 0,},
};

MODULE_DEVICE_TABLE(pci, kfir_pci_tbl);
	
static struct pci_driver kfir_driver = {
	name:		"kfir",
	id_table:	kfir_pci_tbl,
	probe:		kfir_init_one,
	remove:		kfir_remove_one,
	suspend:	kfir_suspend,
	resume:		kfir_resume,
};

int __init kfir_init_module(void)
{
	return pci_module_init(&kfir_driver);
}

void __exit kfir_cleanup_module(void)
{
	pci_unregister_driver(&kfir_driver);
}

module_init(kfir_init_module);
module_exit(kfir_cleanup_module);

#endif 

/*
 * Local variables:
 *  c-indent-level: 8
 *  c-basic-offset: 8
 *  tab-width: 8
 * End:
 */
